"""Obsidian vault representation."""

import csv
import re
import shutil
from dataclasses import dataclass
from pathlib import Path
import json
import rich.repr
from rich import box
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm
from rich.table import Table

from obsidian_metadata._config.config import Config, VaultConfig
from obsidian_metadata._utils import alerts
from obsidian_metadata._utils.alerts import logger as log
from obsidian_metadata.models import InsertLocation, MetadataType, Note, VaultMetadata


@dataclass
class VaultFilter:
    """Vault filters."""

    path_filter: str = None
    key_filter: str = None
    value_filter: str = None
    tag_filter: str = None


@rich.repr.auto
class Vault:
    """Representation of the Obsidian vault.

    Attributes:
        vault (Path): Path to the vault.
        dry_run (bool): Whether to perform a dry run.
        backup_path (Path): Path to the backup of the vault.
        notes (list[Note]): List of all notes in the vault.
    """

    def __init__(
        self,
        config: VaultConfig,
        dry_run: bool = False,
        filters: list[VaultFilter] = [],
    ):
        self.config = config.config
        self.vault_path: Path = config.path
        self.name = self.vault_path.name
        self.insert_location: InsertLocation = self._find_insert_location()
        self.dry_run: bool = dry_run
        self.backup_path: Path = self.vault_path.parent / f"{self.vault_path.name}.bak"
        self.exclude_paths: list[Path] = []
        self.metadata = VaultMetadata()
        for p in config.exclude_paths:
            self.exclude_paths.append(Path(self.vault_path / p))

        self.filters = filters
        self.all_note_paths = self._find_markdown_notes()

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task(description="Processing notes...", total=None)
            self.all_notes: list[Note] = [
                Note(note_path=p, dry_run=self.dry_run) for p in self.all_note_paths
            ]
            self.notes_in_scope = self._filter_notes()

        self._rebuild_vault_metadata()

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Define rich representation of Vault."""
        yield "vault_path", self.vault_path
        yield "dry_run", self.dry_run
        yield "backup_path", self.backup_path
        yield "num_notes", len(self.all_notes)
        yield "num_notes_in_scope", len(self.notes_in_scope)
        yield "exclude_paths", self.exclude_paths

    def _filter_notes(self) -> list[Note]:
        """Filter notes by path and metadata using the filters defined in self.filters.

        Returns:
            list[Note]: List of notes matching the filters.
        """
        notes_list = self.all_notes.copy()

        for _filter in self.filters:
            if _filter.path_filter is not None:
                notes_list = [
                    n
                    for n in notes_list
                    if re.search(_filter.path_filter, str(n.note_path.relative_to(self.vault_path)))
                ]

            if _filter.tag_filter is not None:
                notes_list = [n for n in notes_list if n.contains_inline_tag(_filter.tag_filter)]

            if _filter.key_filter is not None and _filter.value_filter is not None:
                notes_list = [
                    n
                    for n in notes_list
                    if n.contains_metadata(_filter.key_filter, _filter.value_filter)
                ]
            if _filter.key_filter is not None and _filter.value_filter is None:
                notes_list = [n for n in notes_list if n.contains_metadata(_filter.key_filter)]

        return notes_list

    def _find_insert_location(self) -> InsertLocation:
        """Find the insert location for a note.

        Returns:
            InsertLocation: Insert location for the note.
        """
        if self.config["insert_location"].upper() == "TOP":
            return InsertLocation.TOP
        elif self.config["insert_location"].upper() == "HEADER":
            return InsertLocation.AFTER_TITLE
        elif self.config["insert_location"].upper() == "BOTTOM":
            return InsertLocation.BOTTOM
        else:
            return InsertLocation.BOTTOM

    def _find_markdown_notes(self) -> list[Path]:
        """Build list of all markdown files in the vault.

        Returns:
            list[Path]: List of paths to all matching files in the vault.

        """
        return [
            p.resolve()
            for p in self.vault_path.glob("**/*")
            if p.suffix in [".md", ".MD", ".markdown", ".MARKDOWN"]
            and not any(item in p.parents for item in self.exclude_paths)
        ]

    def _rebuild_vault_metadata(self) -> None:
        """Rebuild vault metadata."""
        self.metadata = VaultMetadata()
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task(description="Processing notes...", total=None)
            for _note in self.notes_in_scope:
                self.metadata.index_metadata(
                    area=MetadataType.FRONTMATTER, metadata=_note.frontmatter.dict
                )
                self.metadata.index_metadata(
                    area=MetadataType.INLINE, metadata=_note.inline_metadata.dict
                )
                self.metadata.index_metadata(
                    area=MetadataType.TAGS,
                    metadata=_note.inline_tags.list,
                )

    def add_metadata(
        self,
        area: MetadataType,
        key: str = None,
        value: str | list[str] = None,
        location: InsertLocation = None,
    ) -> int:
        """Add metadata to all notes in the vault which do not already contain it.

        Args:
            area (MetadataType): Area of metadata to add to.
            key (str): Key to add.
            value (str|list, optional): Value to add.
            location (InsertLocation, optional): Location to insert metadata. (Defaults to `vault.config.insert_location`)

        Returns:
            int: Number of notes updated.
        """
        if location is None:
            location = self.insert_location

        num_changed = 0

        for _note in self.notes_in_scope:
            if _note.add_metadata(area=area, key=key, value=value, location=location):
                num_changed += 1

        if num_changed > 0:
            self._rebuild_vault_metadata()

        return num_changed

    def backup(self) -> None:
        """Backup the vault."""
        log.debug("Backing up vault")
        if self.dry_run:
            alerts.dryrun(f"Backup up vault to: {self.backup_path}")
            print("\n")
            return

        try:
            shutil.copytree(self.vault_path, self.backup_path)

        except FileExistsError:  # pragma: no cover
            log.debug("Backup already exists")
            if not Confirm.ask("Vault backup already exists. Overwrite?"):
                alerts.info("Exiting backup not overwritten.")
                return

            log.debug("Overwriting backup")
            shutil.rmtree(self.backup_path)
            shutil.copytree(self.vault_path, self.backup_path)

        alerts.success(f"Vault backed up to: {self.backup_path}")

    def commit_changes(self) -> None:
        """Commit changes by writing to disk."""
        log.debug("Writing changes to vault...")
        if self.dry_run:
            for _note in self.notes_in_scope:
                if _note.has_changes():
                    alerts.dryrun(
                        f"writing changes to {_note.note_path.relative_to(self.vault_path)}"
                    )
            return

        for _note in self.notes_in_scope:
            if _note.has_changes():
                log.trace(f"writing to {_note.note_path}")
                _note.write()

    def delete_backup(self) -> None:
        """Delete the vault backup."""
        log.debug("Deleting vault backup")
        if self.backup_path.exists() and self.dry_run is False:
            shutil.rmtree(self.backup_path)
            alerts.success("Backup deleted")
        elif self.backup_path.exists() and self.dry_run is True:
            alerts.dryrun("Delete backup")
        else:
            alerts.info("No backup found")

    def delete_inline_tag(self, tag: str) -> int:
        """Delete an inline tag in the vault.

        Args:
            tag (str): Tag to delete.

        Returns:
            int: Number of notes that had tag deleted.
        """
        num_changed = 0

        for _note in self.notes_in_scope:
            if _note.delete_inline_tag(tag):
                num_changed += 1

        if num_changed > 0:
            self._rebuild_vault_metadata()

        return num_changed

    def delete_metadata(self, key: str, value: str = None) -> int:
        """Delete metadata in the vault.

        Args:
            key (str): Key to delete. Regex is supported
            value (str, optional): Value to delete. Regex is supported

        Returns:
            int: Number of notes that had metadata deleted.
        """
        num_changed = 0

        for _note in self.notes_in_scope:
            if _note.delete_metadata(key, value):
                num_changed += 1

        if num_changed > 0:
            self._rebuild_vault_metadata()

        return num_changed

    def export_metadata(self, path: str, format: str = "csv") -> None:
        """Write metadata to a csv file.

        Args:
            path (Path): Path to write csv file to.
            export_as (str, optional): Export as 'csv' or 'json'. Defaults to "csv".
        """
        export_file = Path(path).expanduser().resolve()

        match format:  # noqa: E999
            case "csv":
                with open(export_file, "w", encoding="UTF8") as f:
                    writer = csv.writer(f)
                    writer.writerow(["Metadata Type", "Key", "Value"])

                    for key, value in self.metadata.frontmatter.items():
                        if isinstance(value, list):
                            if len(value) > 0:
                                for v in value:
                                    writer.writerow(["frontmatter", key, v])
                            else:
                                writer.writerow(["frontmatter", key, v])

                    for key, value in self.metadata.inline_metadata.items():
                        if isinstance(value, list):
                            if len(value) > 0:
                                for v in value:
                                    writer.writerow(["inline_metadata", key, v])
                            else:
                                writer.writerow(["frontmatter", key, v])
                    for tag in self.metadata.tags:
                        writer.writerow(["tags", "", f"{tag}"])
            case "json":
                dict_to_dump = {
                    "frontmatter": self.metadata.dict,
                    "inline_metadata": self.metadata.inline_metadata,
                    "tags": self.metadata.tags,
                }

                with open(export_file, "w", encoding="UTF8") as f:
                    json.dump(dict_to_dump, f, indent=4, ensure_ascii=False, sort_keys=True)

    def get_changed_notes(self) -> list[Note]:
        """Returns a list of notes that have changes.

        Returns:
            list[Note]: List of notes that have changes.
        """
        changed_notes = []
        for _note in self.notes_in_scope:
            if _note.has_changes():
                changed_notes.append(_note)

        changed_notes = sorted(changed_notes, key=lambda x: x.note_path)
        return changed_notes

    def info(self) -> None:
        """Print information about the vault."""
        table = Table(show_header=False)
        table.add_row("Vault", str(self.vault_path))
        if self.backup_path.exists():
            table.add_row("Backup path", str(self.backup_path))
        else:
            table.add_row("Backup", "None")
        table.add_row("Notes in scope", str(len(self.notes_in_scope)))
        table.add_row("Notes excluded from scope", str(self.num_excluded_notes()))
        table.add_row("Active filters", str(len(self.filters)))
        table.add_row("Notes with changes", str(len(self.get_changed_notes())))

        Console().print(table)

    def list_editable_notes(self) -> None:
        """Print a list of notes within the scope that are being edited."""
        table = Table(title="Notes in current scope", show_header=False, box=box.HORIZONTALS)
        for _n, _note in enumerate(self.notes_in_scope, start=1):
            table.add_row(str(_n), str(_note.note_path.relative_to(self.vault_path)))
        Console().print(table)

    def num_excluded_notes(self) -> int:
        """Count number of excluded notes."""
        return len(self.all_notes) - len(self.notes_in_scope)

    def rename_inline_tag(self, old_tag: str, new_tag: str) -> int:
        """Rename an inline tag in the vault.

        Args:
            old_tag (str): Old tag name.
            new_tag (str): New tag name.

        Returns:
            int: Number of notes that had inline tags renamed.
        """
        num_changed = 0

        for _note in self.notes_in_scope:
            if _note.rename_inline_tag(old_tag, new_tag):
                num_changed += 1

        if num_changed > 0:
            self._rebuild_vault_metadata()

        return num_changed

    def rename_metadata(self, key: str, value_1: str, value_2: str = None) -> int:
        """Renames a key or key-value pair in the note's metadata.

        If no value is provided, will rename an entire key.

        Args:
            key (str): Key to rename.
            value_1 (str): Value to rename or new name of key if no value_2 is provided.
            value_2 (str, optional): New value.

        Returns:
            int: Number of notes that had metadata renamed.
        """
        num_changed = 0

        for _note in self.notes_in_scope:
            if _note.rename_metadata(key, value_1, value_2):
                num_changed += 1

        if num_changed > 0:
            self._rebuild_vault_metadata()

        return num_changed

    def transpose_metadata(
        self,
        begin: MetadataType,
        end: MetadataType,
        key: str = None,
        value: str | list[str] = None,
        location: InsertLocation = None,
    ) -> int:
        """Transpose metadata from one type to another.

        Args:
            begin (MetadataType): Metadata type to transpose from.
            end (MetadataType): Metadata type to transpose to.
            key (str, optional): Key to transpose. Defaults to None.
            value (str, optional): Value to transpose. Defaults to None.
            location (InsertLocation, optional): Location to insert metadata. (Defaults to `vault.config.insert_location`)

        Returns:
            int: Number of notes that had metadata transposed.
        """
        if location is None:
            location = self.insert_location

        num_changed = 0
        for _note in self.notes_in_scope:
            if _note.transpose_metadata(
                begin=begin,
                end=end,
                key=key,
                value=value,
                location=location,
            ):
                num_changed += 1

        if num_changed > 0:
            self._rebuild_vault_metadata()

        return num_changed



"""Representation of notes and in the vault."""


import difflib
import re
from pathlib import Path
import copy
import rich.repr
import typer
from rich.console import Console
from rich.table import Table
from obsidian_metadata._utils import alerts
from obsidian_metadata._utils.alerts import logger as log
from obsidian_metadata.models import (
    Frontmatter,
    InlineMetadata,
    InlineTags,
    InsertLocation,
    MetadataType,
    Patterns,
)

PATTERNS = Patterns()


@rich.repr.auto
class Note:
    """Representation of a note in the vault.

    Args:
        note_path (Path): Path to the note file.

    Attributes:
        note_path (Path): Path to the note file.
        dry_run (bool): Whether to run in dry-run mode.
        file_content (str): Total contents of the note file (frontmatter and content).
        frontmatter (dict): Frontmatter of the note.
        inline_tags (list): List of inline tags in the note.
        inline_metadata (dict): Dictionary of inline metadata in the note.
    """

    def __init__(self, note_path: Path, dry_run: bool = False):
        log.trace(f"Creating Note object for {note_path}")
        self.note_path: Path = Path(note_path)
        self.dry_run: bool = dry_run

        try:
            with self.note_path.open():
                self.file_content: str = self.note_path.read_text()
        except FileNotFoundError as e:
            alerts.error(f"Note {self.note_path} not found. Exiting")
            raise typer.Exit(code=1) from e

        self.frontmatter: Frontmatter = Frontmatter(self.file_content)
        self.inline_tags: InlineTags = InlineTags(self.file_content)
        self.inline_metadata: InlineMetadata = InlineMetadata(self.file_content)
        self.original_file_content: str = self.file_content

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Define rich representation of Vault."""
        yield "note_path", self.note_path
        yield "dry_run", self.dry_run
        yield "frontmatter", self.frontmatter
        yield "inline_tags", self.inline_tags
        yield "inline_metadata", self.inline_metadata

    def _delete_inline_metadata(self, key: str, value: str = None) -> None:
        """Delete an inline metadata key/value pair from the text of the note. This method does not remove the key/value from the metadata attribute of the note.

        Args:
            key (str): Key to delete.
            value (str, optional): Value to delete.
        """
        all_results = PATTERNS.find_inline_metadata.findall(self.file_content)
        stripped_null_values = [tuple(filter(None, x)) for x in all_results]

        for _k, _v in stripped_null_values:
            if re.search(key, _k):
                if value is None:
                    _k = re.escape(_k)
                    _v = re.escape(_v)
                    self.sub(rf"\[?{_k}:: ?{_v}]?", "", is_regex=True)
                    return

                if re.search(value, _v):
                    _k = re.escape(_k)
                    _v = re.escape(_v)
                    self.sub(rf"({_k}::) ?{_v}", r"\1", is_regex=True)

    def _rename_inline_metadata(self, key: str, value_1: str, value_2: str = None) -> None:
        """Replace the inline metadata in the note with the current inline metadata object.

        Args:
            key (str): Key to rename.
            value_1 (str): Value to replace OR new key name (if value_2 is None).
            value_2 (str, optional): New value.

        """
        all_results = PATTERNS.find_inline_metadata.findall(self.file_content)
        stripped_null_values = [tuple(filter(None, x)) for x in all_results]

        for _k, _v in stripped_null_values:
            if re.search(key, _k):
                if value_2 is None:
                    if re.search(rf"{key}[^\w\d_-]+", _k):
                        key_text = re.split(r"[^\w\d_-]+$", _k)[0]
                        key_markdown = re.split(r"^[\w\d_-]+", _k)[1]
                        self.sub(
                            rf"{key_text}{key_markdown}::",
                            rf"{value_1}{key_markdown}::",
                        )
                    else:
                        self.sub(f"{_k}::", f"{value_1}::")
                else:
                    if re.search(key, _k) and re.search(value_1, _v):
                        _k = re.escape(_k)
                        _v = re.escape(_v)
                        self.sub(f"{_k}:: ?{_v}", f"{_k}:: {value_2}", is_regex=True)

    def add_metadata(
        self,
        area: MetadataType,
        key: str = None,
        value: str | list[str] = None,
        location: InsertLocation = None,
    ) -> bool:
        """Add metadata to the note if it does not already exist.

        Args:
            area (MetadataType): Area to add metadata to.
            key (str, optional): Key to add
            location (InsertLocation, optional): Location to add inline metadata and tags.
            value (str, optional): Value to add.

        Returns:
            bool: Whether the metadata was added.
        """
        match area:  # noqa: E999
            case MetadataType.FRONTMATTER if self.frontmatter.add(key, value):
                self.update_frontmatter()
                return True

            case MetadataType.INLINE:
                if value is None:
                    if self.inline_metadata.add(key):
                        line = f"{key}::"
                        self.insert(new_string=line, location=location)
                        return True

                new_values = []
                if isinstance(value, list):
                    new_values = [_v for _v in value if self.inline_metadata.add(key, _v)]
                else:
                    if self.inline_metadata.add(key, value):
                        new_values = [value]

                if new_values:
                    for value in new_values:
                        self.insert(new_string=f"{key}:: {value}", location=location)
                    return True

            case MetadataType.TAGS:
                new_values = []
                if isinstance(value, list):
                    new_values = [_v for _v in value if self.inline_tags.add(_v)]
                else:
                    if self.inline_tags.add(value):
                        new_values = [value]

                if new_values:
                    for value in new_values:
                        if value.startswith("#"):
                            value = value[1:]
                        self.insert(new_string=f"#{value}", location=location)
                    return True

            case _:
                return False

        return False

    def contains_inline_tag(self, tag: str, is_regex: bool = False) -> bool:
        """Check if a note contains the specified inline tag.

        Args:
            tag (str): Tag to check for.
            is_regex (bool, optional): Whether to use regex to match the tag.

        Returns:
            bool: Whether the note has inline tags.
        """
        return self.inline_tags.contains(tag, is_regex=is_regex)

    def contains_metadata(self, key: str, value: str = None, is_regex: bool = False) -> bool:
        """Check if a note has a key or a key-value pair in its metadata.

        Args:
            key (str): Key to check for.
            value (str, optional): Value to check for.
            is_regex (bool, optional): Whether to use regex to match the key/value.

        Returns:
            bool: Whether the note contains the key or key-value pair.
        """
        if value is None:
            if self.frontmatter.contains(key, is_regex=is_regex) or self.inline_metadata.contains(
                key, is_regex=is_regex
            ):
                return True
            return False

        if self.frontmatter.contains(
            key, value, is_regex=is_regex
        ) or self.inline_metadata.contains(key, value, is_regex=is_regex):
            return True

        return False

    def delete_inline_tag(self, tag: str) -> bool:
        """Delete an inline tag from the `inline_tags` attribute AND removes the tag from the text of the note if it exists.

        Args:
            tag (str): Tag to delete.

        Returns:
            bool: Whether the tag was deleted.
        """
        new_list = self.inline_tags.list.copy()

        for _t in new_list:
            if re.search(tag, _t):
                _t = re.escape(_t)
                self.sub(rf"#{_t}([ \|,;:\*\(\)\[\]\\\.\n#&])", r"\1", is_regex=True)
                self.inline_tags.delete(tag)

        if new_list != self.inline_tags.list:
            return True

        return False

    def delete_metadata(
        self, key: str, value: str = None, area: MetadataType = MetadataType.ALL
    ) -> bool:
        """Delete a key or key-value pair from the note's metadata. Regex is supported.

        If no value is provided, will delete an entire key.

        Args:
            key (str): Key to delete.
            value (str, optional): Value to delete.
            area (MetadataType, optional): Area to delete metadata from. Defaults to MetadataType.ALL.

        Returns:
            bool: Whether the key or key-value pair was deleted.
        """
        changed_value: bool = False

        if value is None:
            if (
                area == MetadataType.FRONTMATTER or area == MetadataType.ALL
            ) and self.frontmatter.delete(key):
                self.update_frontmatter()
                changed_value = True
            if (
                area == MetadataType.INLINE or area == MetadataType.ALL
            ) and self.inline_metadata.delete(key):
                self._delete_inline_metadata(key, value)
                changed_value = True
        else:
            if (
                area == MetadataType.FRONTMATTER or area == MetadataType.ALL
            ) and self.frontmatter.delete(key, value):
                self.update_frontmatter()
                changed_value = True
            if (
                area == MetadataType.INLINE or area == MetadataType.ALL
            ) and self.inline_metadata.delete(key, value):
                self._delete_inline_metadata(key, value)
                changed_value = True

        if changed_value:
            return True
        return False

    def has_changes(self) -> bool:
        """Check if the note has been updated.

        Returns:
            bool: Whether the note has been updated.
        """
        if self.frontmatter.has_changes():
            return True

        if self.inline_tags.has_changes():
            return True

        if self.inline_metadata.has_changes():
            return True

        if self.file_content != self.original_file_content:
            return True

        return False

    def insert(
        self,
        new_string: str,
        location: InsertLocation,
        allow_multiple: bool = False,
    ) -> None:
        """Insert a string at the top of a note.

        Args:
            new_string (str): String to insert at the top of the note.
            allow_multiple (bool): Whether to allow inserting the string if it already exists in the note.
            location (InsertLocation): Location to insert the string.
        """
        if not allow_multiple and len(re.findall(re.escape(new_string), self.file_content)) > 0:
            return

        match location:
            case InsertLocation.BOTTOM:
                self.file_content += f"\n{new_string}"
            case InsertLocation.TOP:
                try:
                    top = PATTERNS.frontmatter_block.search(self.file_content).group("frontmatter")
                except AttributeError:
                    top = ""

                if top == "":
                    self.file_content = f"{new_string}\n{self.file_content}"
                else:
                    new_string = f"{top}\n{new_string}"
                    top = re.escape(top)
                    self.sub(top, new_string, is_regex=True)
            case InsertLocation.AFTER_TITLE:
                try:
                    top = PATTERNS.top_with_header.search(self.file_content).group("top")
                except AttributeError:
                    top = ""

                if top == "":
                    self.file_content = f"{new_string}\n{self.file_content}"
                else:
                    new_string = f"{top}\n{new_string}"
                    top = re.escape(top)
                    self.sub(top, new_string, is_regex=True)
            case _:
                raise ValueError(f"Invalid location: {location}")
        pass

    def print_note(self) -> None:
        """Print the note to the console."""
        print(self.file_content)

    def print_diff(self) -> None:
        """Print a diff of the note's original state and it's new state."""
        a = self.original_file_content.splitlines()
        b = self.file_content.splitlines()

        diff = difflib.Differ()
        result = list(diff.compare(a, b))
        table = Table(title=f"\nDiff of {self.note_path.name}", show_header=False, min_width=50)

        for line in result:
            if line.startswith("+"):
                table.add_row(line, style="green")
            elif line.startswith("-"):
                table.add_row(line, style="red")

        Console().print(table)

    def rename_inline_tag(self, tag_1: str, tag_2: str) -> bool:
        """Rename an inline tag from the note ONLY if it's not in the metadata as well.

        Args:
            tag_1 (str): Tag to rename.
            tag_2 (str): New tag name.

        Returns:
            bool: Whether the tag was renamed.
        """
        if tag_1 in self.inline_tags.list:
            self.sub(
                rf"#{tag_1}([ \|,;:\*\(\)\[\]\\\.\n#&])",
                rf"#{tag_2}\1",
                is_regex=True,
            )
            self.inline_tags.rename(tag_1, tag_2)
            return True
        return False

    def rename_metadata(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Rename a key or key-value pair in the note's metadata.

        If no value is provided, will rename an entire key.

        Args:
            key (str): Key to rename.
            value_1 (str): Value to rename or new name of key if no value_2 is provided.
            value_2 (str, optional): New value.

        Returns:
            bool: Whether the note was updated.
        """
        changed_value: bool = False
        if value_2 is None:
            if self.frontmatter.rename(key, value_1):
                self.update_frontmatter()
                changed_value = True
            if self.inline_metadata.rename(key, value_1):
                self._rename_inline_metadata(key, value_1)
                changed_value = True
        else:
            if self.frontmatter.rename(key, value_1, value_2):
                self.update_frontmatter()
                changed_value = True
            if self.inline_metadata.rename(key, value_1, value_2):
                self._rename_inline_metadata(key, value_1, value_2)
                changed_value = True

        if changed_value:
            return True

        return False

    def sub(self, pattern: str, replacement: str, is_regex: bool = False) -> None:
        """Substitutes text within the note.

        Args:
            pattern (str): The pattern to replace (plain text or regular expression).
            replacement (str): What to replace the pattern with.
            is_regex (bool): Whether the pattern is a regex pattern or plain text.
        """
        if not is_regex:
            pattern = re.escape(pattern)

        self.file_content = re.sub(pattern, replacement, self.file_content, re.MULTILINE)

    def transpose_metadata(
        self,
        begin: MetadataType,
        end: MetadataType,
        key: str = None,
        value: str | list[str] = None,
        location: InsertLocation = InsertLocation.BOTTOM,
    ) -> bool:
        """Transpose metadata from one type to another.

        Args:
            begin (MetadataType): The type of metadata to transpose from.
            end (MetadataType): The type of metadata to transpose to.
            key (str, optional): The key to transpose. Defaults to None.
            value (str | list[str], optional): The value to transpose. Defaults to None.

        Returns:
            bool: Whether the note was updated.
        """
        if (begin == MetadataType.FRONTMATTER or begin == MetadataType.INLINE) and (
            end == MetadataType.FRONTMATTER or end == MetadataType.INLINE
        ):
            if begin == MetadataType.FRONTMATTER:
                begin_dict = self.frontmatter.dict
            else:
                begin_dict = self.inline_metadata.dict

            if begin_dict == {}:
                return False

            if key is None:  # Transpose all metadata when no key is provided
                for _key, _value in begin_dict.items():
                    self.add_metadata(key=_key, value=_value, area=end, location=location)
                    self.delete_metadata(key=_key, area=begin)
                return True

            has_changes = False
            temp_dict = copy.deepcopy(begin_dict)
            for k, v in begin_dict.items():
                if key == k:
                    if value is None:
                        self.add_metadata(key=k, value=v, area=end, location=location)
                        self.delete_metadata(key=k, area=begin)
                        return True

                    if value == v:
                        self.add_metadata(key=k, value=v, area=end, location=location)
                        self.delete_metadata(key=k, area=begin)
                        return True

                    if isinstance(value, str):
                        if value in v:
                            self.add_metadata(key=k, value=value, area=end, location=location)
                            self.delete_metadata(key=k, value=value, area=begin)
                            return True
                        else:
                            return False

                    if isinstance(value, list):
                        for value_item in value:
                            if value_item in v:
                                self.add_metadata(
                                    key=k, value=value_item, area=end, location=location
                                )
                                self.delete_metadata(key=k, value=value_item, area=begin)
                                temp_dict[k].remove(value_item)
                                has_changes = True

                        if temp_dict[k] == []:
                            self.delete_metadata(key=k, area=begin)

                    if has_changes:
                        return True
                    else:
                        return False

        if begin == MetadataType.TAGS:
            # TODO: Implement transposing to and from tags
            pass

        return False

    def update_frontmatter(self, sort_keys: bool = False) -> None:
        """Replace the frontmatter in the note with the current frontmatter object."""
        try:
            current_frontmatter = PATTERNS.frontmatter_block.search(self.file_content).group(
                "frontmatter"
            )
        except AttributeError:
            current_frontmatter = None

        if current_frontmatter is None and self.frontmatter.dict == {}:
            return

        new_frontmatter = self.frontmatter.to_yaml(sort_keys=sort_keys)
        if self.frontmatter.dict == {}:
            new_frontmatter = ""
        else:
            new_frontmatter = f"---\n{new_frontmatter}---\n"

        if current_frontmatter is None:
            self.file_content = new_frontmatter + self.file_content
            return

        current_frontmatter = f"{re.escape(current_frontmatter)}\n?"
        self.sub(current_frontmatter, new_frontmatter, is_regex=True)

    def write(self, path: Path = None) -> None:
        """Write the note's content to disk.

        Args:
            path (Path): Path to write the note to. Defaults to the note's path.

        Raises:
            typer.Exit: If the note's path is not found.
        """
        p = self.note_path if path is None else path
        if self.dry_run:
            log.trace(f"DRY RUN: Writing note {p} to disk")
            return

        try:
            with open(p, "w") as f:
                log.trace(f"Writing note {p} to disk")
                f.write(self.file_content)
        except FileNotFoundError as e:
            alerts.error(f"Note {p} not found. Exiting")
            raise typer.Exit(code=1) from e



"""Utility functions."""
import re
from os import name, system
from typing import Any

import typer

from obsidian_metadata.__version__ import __version__


def clean_dictionary(dictionary: dict[str, Any]) -> dict[str, Any]:
    """Clean up a dictionary by markdown formatting from keys and values.

    Args:
        dictionary (dict): Dictionary to clean

    Returns:
        dict: Cleaned dictionary
    """
    new_dict = {key.strip(): value for key, value in dictionary.items()}
    new_dict = {key.strip("*[]#"): value for key, value in new_dict.items()}
    for key, value in new_dict.items():
        new_dict[key] = [s.strip("*[]#") for s in value if isinstance(value, list)]

    return new_dict


def clear_screen() -> None:  # pragma: no cover
    """Clear the screen."""
    # for windows
    _ = system("cls") if name == "nt" else system("clear")


def dict_contains(
    dictionary: dict[str, list[str]], key: str, value: str = None, is_regex: bool = False
) -> bool:
    """Check if a dictionary contains a key.

    Args:
        dictionary (dict): Dictionary to check
        key (str): Key to check for
        value (str, optional): Value to check for. Defaults to None.
        is_regex (bool, optional): Whether the key is a regex. Defaults to False.

    Returns:
        bool: Whether the dictionary contains the key
    """
    if value is None:
        if is_regex:
            return any(re.search(key, str(_key)) for _key in dictionary)
        return key in dictionary

    if is_regex:
        found_keys = []
        for _key in dictionary:
            if re.search(key, str(_key)):
                found_keys.append(
                    any(re.search(value, _v) for _v in dictionary[_key]),
                )
        return any(found_keys)

    return key in dictionary and value in dictionary[key]


def dict_values_to_lists_strings(dictionary: dict, strip_null_values: bool = False) -> dict:
    """Convert all values in a dictionary to lists of strings.

    Args:
        dictionary (dict): Dictionary to convert
        strip_null (bool): Whether to strip null values

    Returns:
        dict: Dictionary with all values converted to lists of strings

        {key: sorted(new_dict[key]) for key in sorted(new_dict)}
    """
    new_dict = {}

    if strip_null_values:
        for key, value in dictionary.items():
            if isinstance(value, list):
                new_dict[key] = sorted([str(item) for item in value if item is not None])
            elif isinstance(value, dict):
                new_dict[key] = dict_values_to_lists_strings(value)  # type: ignore[assignment]
            elif value is None or value == "None" or value == "":
                new_dict[key] = []
            else:
                new_dict[key] = [str(value)]

        return new_dict

    for key, value in dictionary.items():
        if isinstance(value, list):
            new_dict[key] = sorted([str(item) for item in value])
        elif isinstance(value, dict):
            new_dict[key] = dict_values_to_lists_strings(value)  # type: ignore[assignment]
        else:
            new_dict[key] = [str(value)]

    return new_dict


def docstring_parameter(*sub: Any) -> Any:
    """Replace variables within docstrings.

    Args:
        sub (Any): Replacement variables

    Usage:
        @docstring_parameter("foo", "bar")
        def foo():
            '''This is a {0} docstring with {1} variables.'''

    """

    def dec(obj: Any) -> Any:
        """Format object."""
        obj.__doc__ = obj.__doc__.format(*sub)
        return obj

    return dec


def merge_dictionaries(dict1: dict, dict2: dict) -> dict:
    """Merge two dictionaries. When the values are lists, they are merged and sorted.

    Args:
        dict1 (dict): First dictionary.
        dict2 (dict): Second dictionary.

    Returns:
        dict: Merged dictionary.
    """
    for k, v in dict2.items():
        if k in dict1:
            if isinstance(v, list):
                dict1[k].extend(v)
        else:
            dict1[k] = v

    for k, v in dict1.items():
        if isinstance(v, list):
            dict1[k] = sorted(set(v))
        elif isinstance(v, dict):  # pragma: no cover
            for kk, vv in v.items():
                if isinstance(vv, list):
                    v[kk] = sorted(set(vv))

    return dict(sorted(dict1.items()))


def remove_markdown_sections(
    text: str,
    strip_codeblocks: bool = False,
    strip_inlinecode: bool = False,
    strip_frontmatter: bool = False,
) -> str:
    """Strip markdown sections from text.

    Args:
        text (str): Text to remove code blocks from
        strip_codeblocks (bool, optional): Strip code blocks. Defaults to False.
        strip_inlinecode (bool, optional): Strip inline code. Defaults to False.
        strip_frontmatter (bool, optional): Strip frontmatter. Defaults to False.

    Returns:
        str: Text without code blocks
    """
    if strip_codeblocks:
        text = re.sub(r"`{3}.*?`{3}", "", text, flags=re.DOTALL)

    if strip_inlinecode:
        text = re.sub(r"`.*?`", "", text)

    if strip_frontmatter:
        text = re.sub(r"^\s*---.*?---", "", text, flags=re.DOTALL)

    return text  # noqa: RET504


def version_callback(value: bool) -> None:
    """Print version and exit."""
    if value:
        print(f"{__package__.split('.')[0]}: v{__version__}")
        raise typer.Exit()



"""obsidian-metadata CLI."""

from pathlib import Path
from typing import Optional

import questionary
import typer
from rich import print

from obsidian_metadata._config import Config
from obsidian_metadata._utils import (
    alerts,
    clear_screen,
    docstring_parameter,
    version_callback,
)
from obsidian_metadata.models import Application

app = typer.Typer(add_completion=False, no_args_is_help=True, rich_markup_mode="rich")

typer.rich_utils.STYLE_HELPTEXT = ""

HELP_TEXT = """
"""


@app.command()
@docstring_parameter(__package__)
def main(
    config_file: Path = typer.Option(
        Path(Path.home() / f".{__package__}.toml"),
        help="Specify a custom path to a configuration file",
        show_default=False,
    ),
    export_csv: Path = typer.Option(
        None,
        help="Exports all metadata to a specified CSV file and exits. (Will overwrite any existing file)",
        show_default=False,
        dir_okay=False,
        file_okay=True,
    ),
    export_json: Path = typer.Option(
        None,
        help="Exports all metadata to a specified JSON file and exits. (Will overwrite any existing file)",
        show_default=False,
        dir_okay=False,
        file_okay=True,
    ),
    vault_path: Path = typer.Option(
        None,
        help="Path to Obsidian vault",
        show_default=False,
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        "-n",
        help="Dry run - don't actually change anything",
    ),
    log_file: Path = typer.Option(
        Path(Path.home() / "logs" / "obsidian_metadata.log"),
        help="Path to log file",
        show_default=True,
        dir_okay=False,
        file_okay=True,
        exists=False,
    ),
    log_to_file: bool = typer.Option(
        False,
        "--log-to-file",
        help="Log to file",
        show_default=True,
    ),
    verbosity: int = typer.Option(
        0,
        "-v",
        "--verbose",
        show_default=False,
        help="""Set verbosity level (0=WARN, 1=INFO, 2=DEBUG, 3=TRACE)""",
        count=True,
    ),
    version: Optional[bool] = typer.Option(
        None, "--version", help="Print version and exit", callback=version_callback, is_eager=True
    ),
) -> None:
    r"""Make batch updates to metadata in an Obsidian vault. No changes are made to the Vault until they are explicitly committed.

    [bold] [/]
    [bold underline]It is strongly recommended that you back up your vault prior to committing changes.[/] This script makes changes directly to the markdown files in your vault. Once the changes are committed, there is no ability to recreate the original information unless you have a backup.  Follow the instructions in the script to create a backup of your vault if needed.  The author of this script is not responsible for any data loss that may occur. Use at your own risk.

    [bold underline]Configuration:[/]
    Configuration is specified in a configuration file. On First run, this file will be created at [tan]~/.{0}.env[/]. Any options specified on the command line will override the configuration file.

    [bold underline]Usage:[/]
    [tan]Obsidian-metadata[/] provides a menu of sub-commands.

    [bold underline]Vault Actions[/]
    Create or delete a backup of your vault.
        • Backup:        Create a backup of the vault.
        • Delete Backup: Delete a backup of the vault.

    [bold underline]Inspect Metadata[/]
    Inspect the metadata in your vault.
        • View all metadata in the vault
        • View all frontmatter
        • View all inline metadata
        • View all inline tags
        • Export all metadata to CSV or JSON file

    [bold underline]Filter Notes in Scope[/]
    Limit the scope of notes to be processed with one or more filters.
        • Path filter (regex):      Limit scope based on the path or filename
        • Metadata filter:          Limit scope based on a key or key/value pair
        • Tag filter:               Limit scope based on an in-text tag
        • List and clear filters:   List all current filters and clear one or all
        • List notes in scope:      List notes that will be processed.

    [bold underline]Add Metadata[/]
    Add new metadata to your vault.
        • Add new metadata to the frontmatter
        • Add new inline metadata - Set `insert_location` in the config to
          control where the new metadata is inserted. (Default: Bottom)
        • Add new inline tag - Set `insert_location` in the config to
          control where the new tag is inserted. (Default: Bottom)

    [bold underline]Rename Metadata[/]
    Rename either a key and all associated values, a specific value within a key. or an in-text tag.
        • Rename a key
        • Rename a value
        • rename an inline tag

    [bold underline]Delete Metadata[/]
    Delete either a key and all associated values, or a specific value.
        • Delete a key and associated values
        • Delete a value from a key
        • Delete an inline tag

    [bold underline]Transpose Metadata[/]
    Move metadata from inline to frontmatter or the reverse.

        • Transpose all metadata - Moves all frontmatter to inline
          metadata, or the reverse
        • Transpose key - Transposes a specific key and all it's values
        • Transpose value- Transpose a specific key:value pair

    [bold underline]Review Changes[/]
    Prior to committing changes, review all changes that will be made.
        • View a diff of the changes that will be made

    [bold underline]Commit Changes[/]
    Write the changes to disk. This step is not undoable.
        • Commit changes to the vault

    """
    # Instantiate logger
    alerts.LoggerManager(  # pragma: no cover
        log_file,
        verbosity,
        log_to_file,
    )

    banner = r"""
   ___  _         _     _ _
  / _ \| |__  ___(_) __| (_) __ _ _ __
 | | | | '_ \/ __| |/ _` | |/ _` | '_ \
 | |_| | |_) \__ \ | (_| | | (_| | | | |
  \___/|_.__/|___/_|\__,_|_|\__,_|_| |_|
 |  \/  | ___| |_ __ _  __| | __ _| |_ __ _
 | |\/| |/ _ \ __/ _` |/ _` |/ _` | __/ _` |
 | |  | |  __/ || (_| | (_| | (_| | || (_| |
 |_|  |_|\___|\__\__,_|\__,_|\__,_|\__\__,_|
"""
    clear_screen()
    print(banner)

    config: Config = Config(config_path=config_file, vault_path=vault_path)
    if len(config.vaults) == 0:
        typer.echo("No vaults configured. Exiting.")
        raise typer.Exit(1)

    if len(config.vaults) == 1:
        application = Application(dry_run=dry_run, config=config.vaults[0])
    else:
        vault_names = [vault.name for vault in config.vaults]
        vault_name = questionary.select(
            "Select a vault to process:",
            choices=vault_names,
        ).ask()
        if vault_name is None:
            raise typer.Exit(code=1)

        vault_to_use = next(vault for vault in config.vaults if vault.name == vault_name)
        application = Application(dry_run=dry_run, config=vault_to_use)

    if export_json is not None:
        path = Path(export_json).expanduser().resolve()
        application.noninteractive_export_json(path)
        raise typer.Exit(code=0)
    if export_csv is not None:
        path = Path(export_json).expanduser().resolve()
        application.noninteractive_export_csv(path)
        raise typer.Exit(code=0)

    application.application_main()


if __name__ == "__main__":
    app()



"""Work with metadata items."""

import re
from io import StringIO
import copy
from rich import print
from rich.columns import Columns
from rich.console import Console
from rich.table import Table
from ruamel.yaml import YAML

from obsidian_metadata._utils.alerts import logger as log
from obsidian_metadata._utils import alerts
from obsidian_metadata._utils import (
    clean_dictionary,
    dict_contains,
    dict_values_to_lists_strings,
    merge_dictionaries,
    remove_markdown_sections,
)
from obsidian_metadata.models import Patterns  # isort: ignore
from obsidian_metadata.models.enums import MetadataType

PATTERNS = Patterns()
INLINE_TAG_KEY: str = "inline_tag"


class VaultMetadata:
    """Representation of all Metadata in the Vault."""

    def __init__(self) -> None:
        self.dict: dict[str, list[str]] = {}
        self.frontmatter: dict[str, list[str]] = {}
        self.inline_metadata: dict[str, list[str]] = {}
        self.tags: list[str] = []

    def __repr__(self) -> str:
        """Representation of all metadata."""
        return str(self.dict)

    def index_metadata(
        self, area: MetadataType, metadata: dict[str, list[str]] | list[str]
    ) -> None:
        """Index pre-existing metadata in the vault. Takes a dictionary as input and merges it with the existing metadata.  Does not overwrite existing keys.

        Args:
            area (MetadataType): Type of metadata.
            metadata (dict): Metadata to add.
        """
        if isinstance(metadata, dict):
            new_metadata = clean_dictionary(metadata)
            self.dict = merge_dictionaries(self.dict, new_metadata)

        if area == MetadataType.FRONTMATTER:
            self.frontmatter = merge_dictionaries(self.frontmatter, new_metadata)

        if area == MetadataType.INLINE:
            self.inline_metadata = merge_dictionaries(self.inline_metadata, new_metadata)

        if area == MetadataType.TAGS and isinstance(metadata, list):
            self.tags.extend(metadata)
            self.tags = sorted({s.strip("#") for s in self.tags})

    def contains(
        self, area: MetadataType, key: str = None, value: str = None, is_regex: bool = False
    ) -> bool:
        """Check if a key and/or a value exists in the metadata.

        Args:
            area (MetadataType): Type of metadata to check.
            key (str, optional): Key to check.
            value (str, optional): Value to check.
            is_regex (bool, optional): Use regex to check. Defaults to False.

        Returns:
            bool: True if the key exists.

        Raises:
            ValueError: Key must be provided when checking for a key's existence.
            ValueError: Value must be provided when checking for a tag's existence.
        """
        if area != MetadataType.TAGS and key is None:
            raise ValueError("Key must be provided when checking for a key's existence.")

        match area:  # noqa: E999
            case MetadataType.ALL:
                if dict_contains(self.dict, key, value, is_regex):
                    return True
                if key is None and value is not None:
                    if is_regex:
                        return any(re.search(value, tag) for tag in self.tags)
                    return value in self.tags

            case MetadataType.FRONTMATTER:
                return dict_contains(self.frontmatter, key, value, is_regex)
            case MetadataType.INLINE:
                return dict_contains(self.inline_metadata, key, value, is_regex)
            case MetadataType.KEYS:
                return dict_contains(self.dict, key, value, is_regex)
            case MetadataType.TAGS:
                if value is None:
                    raise ValueError("Value must be provided when checking for a tag's existence.")
                if is_regex:
                    return any(re.search(value, tag) for tag in self.tags)
                return value in self.tags

        return False

    def delete(self, key: str, value_to_delete: str = None) -> bool:
        """Delete a key or a key's value from the metadata. Regex is supported to allow deleting more than one key or value.

        Args:
            key (str): Key to check.
            value_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_dict = copy.deepcopy(self.dict)

        if value_to_delete is None:
            for _k in list(new_dict):
                if re.search(key, _k):
                    del new_dict[_k]
        else:
            for _k, _v in new_dict.items():
                if re.search(key, _k):
                    new_values = [x for x in _v if not re.search(value_to_delete, x)]
                    new_dict[_k] = sorted(new_values)

        if new_dict != self.dict:
            self.dict = dict(new_dict)
            return True

        return False

    def print_metadata(self, area: MetadataType) -> None:
        """Print metadata to the terminal.

        Args:
            area (MetadataType): Type of metadata to print
        """
        dict_to_print: dict[str, list[str]] = None
        list_to_print: list[str] = None
        match area:
            case MetadataType.INLINE:
                dict_to_print = self.inline_metadata.copy()
                header = "All inline metadata"
            case MetadataType.FRONTMATTER:
                dict_to_print = self.frontmatter.copy()
                header = "All frontmatter"
            case MetadataType.TAGS:
                list_to_print = []
                for tag in self.tags:
                    list_to_print.append(f"#{tag}")
                header = "All inline tags"
            case MetadataType.KEYS:
                list_to_print = sorted(self.dict.keys())
                header = "All Keys"
            case MetadataType.ALL:
                dict_to_print = self.dict.copy()
                list_to_print = []
                for tag in self.tags:
                    list_to_print.append(f"#{tag}")
                header = "All metadata"

        if dict_to_print is not None:
            table = Table(title=header, show_footer=False, show_lines=True)
            table.add_column("Keys")
            table.add_column("Values")
            for key, value in sorted(dict_to_print.items()):
                values: str | dict[str, list[str]] = (
                    "\n".join(sorted(value)) if isinstance(value, list) else value
                )
                table.add_row(f"[bold]{key}[/]", str(values))
            Console().print(table)

        if list_to_print is not None:
            columns = Columns(
                sorted(list_to_print),
                equal=True,
                expand=True,
                title=header if area != MetadataType.ALL else "All inline tags",
            )
            print(columns)

    def rename(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Replace a value in the frontmatter.

        Args:
            key (str): Key to check.
            value_1 (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            value_2 (str, Optional): New value.
            bypass_check (bool, optional): Bypass the check if the key exists. Defaults to False.

        Returns:
            bool: True if a value was renamed
        """
        if value_2 is None:
            if key in self.dict and value_1 not in self.dict:
                self.dict[value_1] = self.dict.pop(key)
                return True
            return False

        if key in self.dict and value_1 in self.dict[key]:
            self.dict[key] = sorted({value_2 if x == value_1 else x for x in self.dict[key]})
            return True

        return False


class Frontmatter:
    """Representation of frontmatter metadata."""

    def __init__(self, file_content: str):
        self.dict: dict[str, list[str]] = self._grab_note_frontmatter(file_content)
        self.dict_original: dict[str, list[str]] = copy.deepcopy(self.dict)

    def __repr__(self) -> str:  # pragma: no cover
        """Representation of the frontmatter.

        Returns:
            str: frontmatter
        """
        return f"Frontmatter(frontmatter={self.dict})"

    def _grab_note_frontmatter(self, file_content: str) -> dict:
        """Grab metadata from a note.

        Args:
            note_path (Path): Path to the note file.

        Returns:
            dict: Metadata from the note.
        """
        try:
            frontmatter_block: str = PATTERNS.frontmatt_block_strip_separators.search(
                file_content
            ).group("frontmatter")
        except AttributeError:
            return {}

        yaml = YAML(typ="safe")
        frontmatter: dict = yaml.load(frontmatter_block)

        for k in frontmatter:
            if frontmatter[k] is None:
                frontmatter[k] = []

        return dict_values_to_lists_strings(frontmatter, strip_null_values=True)

    def add(self, key: str, value: str | list[str] = None) -> bool:
        """Add a key and value to the frontmatter.

        Args:
            key (str): Key to add.
            value (str, optional): Value to add.

        Returns:
            bool: True if the metadata was added
        """
        if value is None:
            if key not in self.dict:
                self.dict[key] = []
                return True
            return False

        if key not in self.dict:
            if isinstance(value, list):
                self.dict[key] = value
                return True

            self.dict[key] = [value]
            return True

        if key in self.dict and value not in self.dict[key]:
            if isinstance(value, list):
                self.dict[key].extend(value)
                return True

            self.dict[key].append(value)
            return True

        return False

    def contains(self, key: str, value: str = None, is_regex: bool = False) -> bool:
        """Check if a key or value exists in the metadata.

        Args:
            key (str): Key to check.
            value (str, optional): Value to check.
            is_regex (bool, optional): Use regex to check. Defaults to False.

        Returns:
            bool: True if the key exists.
        """
        return dict_contains(self.dict, key, value, is_regex)

    def delete(self, key: str, value_to_delete: str = None) -> bool:
        """Delete a value or key in the frontmatter.  Regex is supported to allow deleting more than one key or value.

        Args:
            key (str): If no value, key to delete. If value, key containing the value.
            value_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_dict = dict(self.dict)

        if value_to_delete is None:
            for _k in list(new_dict):
                if re.search(key, _k):
                    del new_dict[_k]
        else:
            for _k, _v in new_dict.items():
                if re.search(key, _k):
                    new_values = [x for x in _v if not re.search(value_to_delete, x)]
                    new_dict[_k] = sorted(new_values)

        if new_dict != self.dict:
            self.dict = dict(new_dict)
            return True

        return False

    def has_changes(self) -> bool:
        """Check if the frontmatter has changes.

        Returns:
            bool: True if the frontmatter has changes.
        """
        return self.dict != self.dict_original

    def rename(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Replace a value in the frontmatter.

        Args:
            key (str): Key to check.
            value_1 (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            value_2 (str, Optional): New value.

        Returns:
            bool: True if a value was renamed
        """
        if value_2 is None:
            if key in self.dict and value_1 not in self.dict:
                self.dict[value_1] = self.dict.pop(key)
                return True
            return False

        if key in self.dict and value_1 in self.dict[key]:
            self.dict[key] = sorted({value_2 if x == value_1 else x for x in self.dict[key]})
            return True

        return False

    def to_yaml(self, sort_keys: bool = False) -> str:
        """Return the frontmatter as a YAML string.

        Returns:
            str: Frontmatter as a YAML string.
            sort_keys (bool, optional): Sort the keys. Defaults to False.
        """
        dict_to_dump = copy.deepcopy(self.dict)
        for k in dict_to_dump:
            if dict_to_dump[k] == []:
                dict_to_dump[k] = None
            if isinstance(dict_to_dump[k], list) and len(dict_to_dump[k]) == 1:
                new_val = dict_to_dump[k][0]
                dict_to_dump[k] = new_val  # type: ignore [assignment]

        # Converting stream to string from https://stackoverflow.com/questions/47614862/best-way-to-use-ruamel-yaml-to-dump-yaml-to-string-not-to-stream/63179923#63179923

        if sort_keys:
            dict_to_dump = dict(sorted(dict_to_dump.items()))

        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        string_stream = StringIO()
        yaml.dump(dict_to_dump, string_stream)
        yaml_value = string_stream.getvalue()
        string_stream.close()
        return yaml_value


class InlineMetadata:
    """Representation of inline metadata in the form of `key:: value`."""

    def __init__(self, file_content: str):
        self.dict: dict[str, list[str]] = self.grab_inline_metadata(file_content)
        self.dict_original: dict[str, list[str]] = copy.deepcopy(self.dict)

    def __repr__(self) -> str:  # pragma: no cover
        """Representation of inline metadata.

        Returns:
            str: inline metadata
        """
        return f"InlineMetadata(inline_metadata={self.dict})"

    def add(self, key: str, value: str | list[str] = None) -> bool:
        """Add a key and value to the inline metadata.

        Args:
            key (str): Key to add.
            value (str, optional): Value to add.

        Returns:
            bool: True if the metadata was added
        """
        if value is None:
            if key not in self.dict:
                self.dict[key] = []
                return True
            return False

        if key not in self.dict:
            if isinstance(value, list):
                self.dict[key] = value
                return True

            self.dict[key] = [value]
            return True

        if key in self.dict and value not in self.dict[key]:
            if isinstance(value, list):
                self.dict[key].extend(value)
                return True

            self.dict[key].append(value)
            return True

        return False

    def contains(self, key: str, value: str = None, is_regex: bool = False) -> bool:
        """Check if a key or value exists in the inline metadata.

        Args:
            key (str): Key to check.
            value (str, Optional): Value to check.
            is_regex (bool, optional): If True, key and value are treated as regex. Defaults to False.

        Returns:
            bool: True if the key exists.
        """
        return dict_contains(self.dict, key, value, is_regex)

    def delete(self, key: str, value_to_delete: str = None) -> bool:
        """Delete a value or key in the inline metadata. Regex is supported to allow deleting more than one key or value.

        Args:
            key (str): If no value, key to delete. If value, key containing the value.
            value_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_dict = dict(self.dict)

        if value_to_delete is None:
            for _k in list(new_dict):
                if re.search(key, _k):
                    del new_dict[_k]
        else:
            for _k, _v in new_dict.items():
                if re.search(key, _k):
                    new_values = [x for x in _v if not re.search(value_to_delete, x)]
                    new_dict[_k] = sorted(new_values)

        if new_dict != self.dict:
            self.dict = dict(new_dict)
            return True

        return False

    def grab_inline_metadata(self, file_content: str) -> dict[str, list[str]]:
        """Grab inline metadata from a note.

        Returns:
            dict[str, str]: Inline metadata from the note.
        """
        content = remove_markdown_sections(
            file_content,
            strip_codeblocks=True,
            strip_inlinecode=True,
            strip_frontmatter=True,
        )
        all_results = PATTERNS.find_inline_metadata.findall(content)
        stripped_null_values = [tuple(filter(None, x)) for x in all_results]

        inline_metadata: dict[str, list[str]] = {}
        for k, v in stripped_null_values:
            if k in inline_metadata:
                inline_metadata[k].append(str(v))
            else:
                inline_metadata[k] = [str(v)]

        return clean_dictionary(inline_metadata)

    def has_changes(self) -> bool:
        """Check if the metadata has changes.

        Returns:
            bool: True if the metadata has changes.
        """
        return self.dict != self.dict_original

    def rename(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Replace a value in the inline metadata.

        Args:
            key (str): Key to check.
            value_1 (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            value_2 (str, Optional): New value.

        Returns:
            bool: True if a value was renamed
        """
        if value_2 is None:
            if key in self.dict and value_1 not in self.dict:
                self.dict[value_1] = self.dict.pop(key)
                return True
            return False

        if key in self.dict and value_1 in self.dict[key]:
            self.dict[key] = sorted({value_2 if x == value_1 else x for x in self.dict[key]})
            return True

        return False


class InlineTags:
    """Representation of inline tags."""

    def __init__(self, file_content: str):
        self.metadata_key = INLINE_TAG_KEY
        self.list: list[str] = self._grab_inline_tags(file_content)
        self.list_original: list[str] = self.list.copy()

    def __repr__(self) -> str:  # pragma: no cover
        """Representation of the inline tags.

        Returns:
            str: inline tags
        """
        return f"InlineTags(tags={self.list})"

    def _grab_inline_tags(self, file_content: str) -> list[str]:
        """Grab inline tags from a note.

        Args:
            file_content (str): Total contents of the note file (frontmatter and content).

        Returns:
            list[str]: Inline tags from the note.
        """
        return sorted(
            PATTERNS.find_inline_tags.findall(
                remove_markdown_sections(
                    file_content,
                    strip_codeblocks=True,
                    strip_inlinecode=True,
                )
            )
        )

    def add(self, new_tag: str | list[str]) -> bool:
        """Add a new inline tag.

        Args:
            new_tag (str): Tag to add.

        Returns:
            bool: True if a tag was added.
        """
        if isinstance(new_tag, list):
            for _tag in new_tag:
                if _tag.startswith("#"):
                    _tag = _tag[1:]
                if _tag in self.list:
                    return False
                new_list = self.list.copy()
                new_list.append(_tag)
                self.list = sorted(new_list)
                return True
        else:
            if new_tag.startswith("#"):
                new_tag = new_tag[1:]
            if new_tag in self.list:
                return False
            new_list = self.list.copy()
            new_list.append(new_tag)
            self.list = sorted(new_list)
            return True

        return False

    def contains(self, tag: str, is_regex: bool = False) -> bool:
        """Check if a tag exists in the metadata.

        Args:
            tag (str): Tag to check.
            is_regex (bool, optional): If True, tag is treated as regex. Defaults to False.

        Returns:
            bool: True if the tag exists.
        """
        if is_regex is True:
            return any(re.search(tag, _t) for _t in self.list)

        if tag in self.list:
            return True

        return False

    def delete(self, tag_to_delete: str) -> bool:
        """Delete a specified inline tag. Regex is supported to allow deleting more than one tag.

        Args:
            tag_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_list = sorted([x for x in self.list if re.search(tag_to_delete, x) is None])

        if new_list != self.list:
            self.list = new_list
            return True
        return False

    def has_changes(self) -> bool:
        """Check if the metadata has changes.

        Returns:
            bool: True if the metadata has changes.
        """
        return self.list != self.list_original

    def rename(self, old_tag: str, new_tag: str) -> bool:
        """Replace an inline tag with another string.

        Args:
            old_tag (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            new_tag (str, Optional): New value.

        Returns:
            bool: True if a value was renamed
        """
        if old_tag in self.list:
            self.list = sorted([new_tag if i == old_tag else i for i in self.list])
            return True
        return False



"""Shared models."""
from obsidian_metadata.models.enums import (
    InsertLocation,
    MetadataType,
)

from obsidian_metadata.models.patterns import Patterns  # isort: skip
from obsidian_metadata.models.metadata import (
    Frontmatter,
    InlineMetadata,
    InlineTags,
    VaultMetadata,
)
from obsidian_metadata.models.notes import Note
from obsidian_metadata.models.vault import Vault, VaultFilter

from obsidian_metadata.models.application import Application  # isort: skip

__all__ = [
    "Application",
    "Frontmatter",
    "InlineMetadata",
    "InlineTags",
    "InsertLocation",
    "LoggerManager",
    "MetadataType",
    "Note",
    "Patterns",
    "Vault",
    "VaultFilter",
    "VaultMetadata",
]



"""Questions for the cli."""


from typing import Any
from pathlib import Path
import questionary
from rich import print
from rich import box
from rich.console import Console
from rich.table import Table
from obsidian_metadata._config import VaultConfig
from obsidian_metadata._utils.alerts import logger as log
from obsidian_metadata.models import Patterns, Vault, VaultFilter
from obsidian_metadata._utils import alerts
from obsidian_metadata.models.questions import Questions
from obsidian_metadata.models.enums import MetadataType

PATTERNS = Patterns()


class Application:
    """Questions for use in the cli.

    Contains methods which ask a series of questions to the user and return a dictionary with their answers.

    More info: https://questionary.readthedocs.io/en/stable/pages/advanced.html#create-questions-from-dictionaries
    """

    def __init__(self, config: VaultConfig, dry_run: bool) -> None:
        self.config = config
        self.dry_run = dry_run
        self.questions = Questions()
        self.filters: list[VaultFilter] = []

    def _load_vault(self) -> None:
        """Load the vault."""

        if len(self.filters) == 0:
            self.vault: Vault = Vault(config=self.config, dry_run=self.dry_run)
        else:
            self.vault = Vault(config=self.config, dry_run=self.dry_run, filters=self.filters)

        alerts.success(
            f"Loaded {len(self.vault.notes_in_scope)} notes from {len(self.vault.all_notes)} total notes"
        )
        self.questions = Questions(vault=self.vault)

    def application_main(self) -> None:
        """Questions for the main application."""
        self._load_vault()

        while True:
            self.vault.info()

            match self.questions.ask_application_main():  # noqa: E999
                case "vault_actions":
                    self.application_vault()
                case "inspect_metadata":
                    self.application_inspect_metadata()
                case "filter_notes":
                    self.application_filter()
                case "add_metadata":
                    self.application_add_metadata()
                case "rename_metadata":
                    self.application_rename_metadata()
                case "delete_metadata":
                    self.application_delete_metadata()
                case "transpose_metadata":
                    self.application_transpose_metadata()
                case "review_changes":
                    self.review_changes()
                case "commit_changes":
                    self.commit_changes()
                case _:
                    break

        print("Done!")
        return

    def application_add_metadata(self) -> None:
        """Add metadata."""
        alerts.usage(
            "Add new metadata to your vault. Currently only supports adding to the frontmatter of a note."
        )

        area = self.questions.ask_area()
        match area:
            case MetadataType.FRONTMATTER | MetadataType.INLINE:
                key = self.questions.ask_new_key(question="Enter the key for the new metadata")
                if key is None:  # pragma: no cover
                    return

                value = self.questions.ask_new_value(
                    question="Enter the value for the new metadata"
                )
                if value is None:  # pragma: no cover
                    return

                num_changed = self.vault.add_metadata(
                    area=area, key=key, value=value, location=self.vault.insert_location
                )
                if num_changed == 0:  # pragma: no cover
                    alerts.warning(f"No notes were changed")
                    return

                alerts.success(f"Added metadata to {num_changed} notes")

            case MetadataType.TAGS:
                tag = self.questions.ask_new_tag()
                if tag is None:  # pragma: no cover
                    return

                num_changed = self.vault.add_metadata(
                    area=area, value=tag, location=self.vault.insert_location
                )

                if num_changed == 0:  # pragma: no cover
                    alerts.warning(f"No notes were changed")
                    return

                alerts.success(f"Added metadata to {num_changed} notes")
            case _:  # pragma: no cover
                return

    def application_delete_metadata(self) -> None:
        alerts.usage("Delete either a key and all associated values, or a specific value.")

        choices = [
            {"name": "Delete key", "value": "delete_key"},
            {"name": "Delete value", "value": "delete_value"},
            {"name": "Delete inline tag", "value": "delete_inline_tag"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        match self.questions.ask_selection(
            choices=choices, question="Select a metadata type to delete"
        ):
            case "delete_key":
                self.delete_key()
            case "delete_value":
                self.delete_value()
            case "delete_inline_tag":
                self.delete_inline_tag()
            case _:  # pragma: no cover
                return

    def application_rename_metadata(self) -> None:
        """Rename metadata."""
        alerts.usage("Select the type of metadata to rename.")

        choices = [
            {"name": "Rename key", "value": "rename_key"},
            {"name": "Rename value", "value": "rename_value"},
            {"name": "Rename inline tag", "value": "rename_inline_tag"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        match self.questions.ask_selection(
            choices=choices, question="Select a metadata type to rename"
        ):
            case "rename_key":
                self.rename_key()
            case "rename_value":
                self.rename_value()
            case "rename_inline_tag":
                self.rename_inline_tag()
            case _:  # pragma: no cover
                return

    def application_filter(self) -> None:
        """Filter notes."""
        alerts.usage("Limit the scope of notes to be processed with one or more filters.")

        choices = [
            {"name": "Apply new regex path filter", "value": "apply_path_filter"},
            {"name": "Apply new metadata filter", "value": "apply_metadata_filter"},
            {"name": "Apply new in-text tag filter", "value": "apply_tag_filter"},
            {"name": "List and clear filters", "value": "list_filters"},
            {"name": "List notes in scope", "value": "list_notes"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        while True:
            match self.questions.ask_selection(choices=choices, question="Select an action"):
                case "apply_path_filter":
                    path = self.questions.ask_filter_path()
                    if path is None or path == "":  # pragma: no cover
                        return

                    self.filters.append(VaultFilter(path_filter=path))
                    self._load_vault()

                case "apply_metadata_filter":
                    key = self.questions.ask_existing_key()
                    if key is None:  # pragma: no cover
                        return

                    questions2 = Questions(vault=self.vault, key=key)
                    value = questions2.ask_existing_value(
                        question="Enter the value for the metadata filter",
                    )
                    if value is None:  # pragma: no cover
                        return
                    if value == "":
                        self.filters.append(VaultFilter(key_filter=key))
                    else:
                        self.filters.append(VaultFilter(key_filter=key, value_filter=value))
                    self._load_vault()

                case "apply_tag_filter":
                    tag = self.questions.ask_existing_inline_tag()
                    if tag is None or tag == "":
                        return

                    self.filters.append(VaultFilter(tag_filter=tag))
                    self._load_vault()

                case "list_filters":
                    if len(self.filters) == 0:
                        alerts.notice("No filters have been applied")
                        return

                    print("")
                    table = Table(
                        "Opt",
                        "Filter",
                        "Type",
                        title="Current Filters",
                        show_header=False,
                        box=box.HORIZONTALS,
                    )
                    for _n, filter in enumerate(self.filters, start=1):
                        if filter.path_filter is not None:
                            table.add_row(
                                str(_n),
                                f"Path regex: [tan bold]{filter.path_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                        elif filter.tag_filter is not None:
                            table.add_row(
                                str(_n),
                                f"Tag filter: [tan bold]{filter.tag_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                        elif filter.key_filter is not None and filter.value_filter is None:
                            table.add_row(
                                str(_n),
                                f"Key filter: [tan bold]{filter.key_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                        elif filter.key_filter is not None and filter.value_filter is not None:
                            table.add_row(
                                str(_n),
                                f"Key/Value : [tan bold]{filter.key_filter}={filter.value_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                    table.add_row(f"{len(self.filters) + 1}", "Clear All")
                    table.add_row(f"{len(self.filters) + 2}", "Return to Main Menu")
                    Console().print(table)

                    num = self.questions.ask_number(
                        question="Enter the number of the filter to clear"
                    )
                    if num is None:
                        return
                    if int(num) <= len(self.filters):
                        self.filters.pop(int(num) - 1)
                        self._load_vault()
                        return
                    if int(num) == len(self.filters) + 1:
                        self.filters = []
                        self._load_vault()
                        return

                case "list_notes":
                    self.vault.list_editable_notes()

                case _:
                    return

    def application_inspect_metadata(self) -> None:
        """View metadata."""
        alerts.usage(
            "Inspect the metadata in your vault. Note, uncommitted changes will be reflected in these reports"
        )

        choices = [
            {"name": "View all metadata", "value": "all_metadata"},
            {"name": "View all frontmatter", "value": "all_frontmatter"},
            {"name": "View all inline_metadata", "value": "all_inline"},
            {"name": "View all keys", "value": "all_keys"},
            {"name": "View all inline tags", "value": "all_tags"},
            questionary.Separator(),
            {"name": "Write all metadata to CSV", "value": "export_csv"},
            {"name": "Write all metadata to JSON file", "value": "export_json"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        while True:
            match self.questions.ask_selection(choices=choices, question="Select a vault action"):
                case "all_metadata":
                    print("")
                    self.vault.metadata.print_metadata(area=MetadataType.ALL)
                    print("")
                case "all_frontmatter":
                    print("")
                    self.vault.metadata.print_metadata(area=MetadataType.FRONTMATTER)
                    print("")
                case "all_inline":
                    print("")
                    self.vault.metadata.print_metadata(area=MetadataType.INLINE)
                    print("")
                case "all_keys":
                    print("")
                    self.vault.metadata.print_metadata(area=MetadataType.KEYS)
                    print("")
                case "all_tags":
                    print("")
                    self.vault.metadata.print_metadata(area=MetadataType.TAGS)
                    print("")
                case "export_csv":
                    path = self.questions.ask_path(question="Enter a path for the CSV file")
                    if path is None:
                        return
                    self.vault.export_metadata(path=path, format="csv")
                    alerts.success(f"Metadata written to {path}")
                case "export_json":
                    path = self.questions.ask_path(question="Enter a path for the JSON file")
                    if path is None:
                        return
                    self.vault.export_metadata(path=path, format="json")
                    alerts.success(f"Metadata written to {path}")
                case _:
                    return

    def application_transpose_metadata(self) -> None:
        """Transpose metadata."""
        alerts.usage("Transpose metadata from frontmatter to inline or vice versa.")

        choices = [
            {"name": "Transpose frontmatter to inline", "value": "frontmatter_to_inline"},
            {"name": "Transpose inline to frontmatter", "value": "inline_to_frontmatter"},
        ]
        match self.questions.ask_selection(
            choices=choices, question="Select metadata to transpose"
        ):
            case "frontmatter_to_inline":
                self.transpose_metadata(begin=MetadataType.FRONTMATTER, end=MetadataType.INLINE)
            case "inline_to_frontmatter":
                self.transpose_metadata(begin=MetadataType.INLINE, end=MetadataType.FRONTMATTER)
            case _:  # pragma: no cover
                return

    def application_vault(self) -> None:
        """Vault actions."""
        alerts.usage("Create or delete a backup of your vault.")

        choices = [
            {"name": "Backup vault", "value": "backup_vault"},
            {"name": "Delete vault backup", "value": "delete_backup"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]

        while True:
            match self.questions.ask_selection(choices=choices, question="Select a vault action"):
                case "backup_vault":
                    self.vault.backup()
                case "delete_backup":
                    self.vault.delete_backup()
                case _:
                    return

    def commit_changes(self) -> bool:
        """Write all changes to disk.

        Returns:
            True if changes were committed, False otherwise.
        """
        changed_notes = self.vault.get_changed_notes()

        if len(changed_notes) == 0:
            print("\n")
            alerts.notice("No changes to commit.\n")
            return False

        backup = questionary.confirm("Create backup before committing changes").ask()
        if backup is None:
            return False
        if backup:
            self.vault.backup()

        if questionary.confirm(f"Commit {len(changed_notes)} changed files to disk?").ask():
            self.vault.commit_changes()

        if not self.dry_run:
            alerts.success(f"{len(changed_notes)} changes committed to disk. Exiting")
            return True

        return True

    def delete_inline_tag(self) -> None:
        """Delete an inline tag."""
        tag = self.questions.ask_existing_inline_tag(question="Which tag would you like to delete?")

        num_changed = self.vault.delete_inline_tag(tag)
        if num_changed == 0:
            alerts.warning(f"No notes were changed")
            return

        alerts.success(f"Deleted inline tag: {tag} in {num_changed} notes")
        return

    def delete_key(self) -> None:
        """Delete a key from the vault."""
        key_to_delete = self.questions.ask_existing_keys_regex(
            question="Regex for the key(s) you'd like to delete?"
        )
        if key_to_delete is None:  # pragma: no cover
            return

        num_changed = self.vault.delete_metadata(key_to_delete)
        if num_changed == 0:
            alerts.warning(f"No notes found with a key matching: [reverse]{key_to_delete}[/]")
            return

        alerts.success(
            f"Deleted keys matching: [reverse]{key_to_delete}[/] from {num_changed} notes"
        )

        return

    def delete_value(self) -> None:
        """Delete a value from the vault."""
        key = self.questions.ask_existing_key(question="Which key contains the value to delete?")
        if key is None:  # pragma: no cover
            return

        questions2 = Questions(vault=self.vault, key=key)
        value = questions2.ask_existing_value_regex(question="Regex for the value to delete")
        if value is None:  # pragma: no cover
            return

        num_changed = self.vault.delete_metadata(key, value)
        if num_changed == 0:
            alerts.warning(f"No notes found matching: {key}: {value}")
            return

        alerts.success(
            f"Deleted value [reverse]{value}[/] from key [reverse]{key}[/] in {num_changed} notes"
        )

        return

    def noninteractive_export_csv(self, path: Path) -> None:
        """Export the vault metadata to CSV."""
        self._load_vault()
        self.vault.export_metadata(format="json", path=str(path))
        alerts.success(f"Exported metadata to {path}")

    def noninteractive_export_json(self, path: Path) -> None:
        """Export the vault metadata to JSON."""
        self._load_vault()
        self.vault.export_metadata(format="json", path=str(path))
        alerts.success(f"Exported metadata to {path}")

    def rename_key(self) -> None:
        """Renames a key in the vault."""

        original_key = self.questions.ask_existing_key(
            question="Which key would you like to rename?"
        )
        if original_key is None:  # pragma: no cover
            return

        new_key = self.questions.ask_new_key()
        if new_key is None:  # pragma: no cover
            return

        num_changed = self.vault.rename_metadata(original_key, new_key)
        if num_changed == 0:
            alerts.warning(f"No notes were changed")
            return

        alerts.success(
            f"Renamed [reverse]{original_key}[/] to [reverse]{new_key}[/] in {num_changed} notes"
        )

    def rename_inline_tag(self) -> None:
        """Rename an inline tag."""

        original_tag = self.questions.ask_existing_inline_tag(question="Which tag to rename?")
        if original_tag is None:  # pragma: no cover
            return

        new_tag = self.questions.ask_new_tag("New tag")
        if new_tag is None:  # pragma: no cover
            return

        num_changed = self.vault.rename_inline_tag(original_tag, new_tag)
        if num_changed == 0:
            alerts.warning(f"No notes were changed")
            return

        alerts.success(
            f"Renamed [reverse]{original_tag}[/] to [reverse]{new_tag}[/] in {num_changed} notes"
        )
        return

    def rename_value(self) -> None:
        """Rename a value in the vault."""
        key = self.questions.ask_existing_key(question="Which key contains the value to rename?")
        if key is None:  # pragma: no cover
            return

        question_key = Questions(vault=self.vault, key=key)
        value = question_key.ask_existing_value(question="Which value would you like to rename?")
        if value is None:  # pragma: no cover
            return

        new_value = question_key.ask_new_value()
        if new_value is None:  # pragma: no cover
            return

        num_changes = self.vault.rename_metadata(key, value, new_value)
        if num_changes == 0:
            alerts.warning(f"No notes were changed")
            return

        alerts.success(f"Renamed '{key}:{value}' to '{key}:{new_value}' in {num_changes} notes")

    def review_changes(self) -> None:
        """Review all changes in the vault."""
        changed_notes = self.vault.get_changed_notes()

        if len(changed_notes) == 0:
            alerts.info("No changes to review.")
            return

        alerts.info(f"Found {len(changed_notes)} changed notes in the vault")
        choices: list[dict[str, Any] | questionary.Separator] = [questionary.Separator()]
        for n, note in enumerate(changed_notes, start=1):
            _selection = {
                "name": f"{n}: {note.note_path.relative_to(self.vault.vault_path)}",
                "value": n - 1,
            }
            choices.append(_selection)

        choices.append(questionary.Separator())
        choices.append({"name": "Return", "value": "return"})

        while True:
            note_to_review = self.questions.ask_selection(
                choices=choices,
                question="Select an updated note to view the diff",
            )
            if note_to_review is None or note_to_review == "return":
                break
            changed_notes[note_to_review].print_diff()

    def transpose_metadata(self, begin: MetadataType, end: MetadataType) -> None:
        """Transpose metadata from one format to another.

        Args:
            begin: The format to transpose from.
            end: The format to transpose to.
        """
        choices = [
            {"name": f"Transpose all {begin.value} to {end.value}", "value": "transpose_all"},
            {"name": "Transpose a key", "value": "transpose_key"},
            {"name": "Transpose a value", "value": "transpose_value"},
            {"name": "Back", "value": "back"},
        ]
        match self.questions.ask_selection(choices=choices, question="Select an action to perform"):
            case "transpose_all":
                num_changed = self.vault.transpose_metadata(
                    begin=begin,
                    end=end,
                    location=self.vault.insert_location,
                )

                if num_changed == 0:
                    alerts.warning(f"No notes were changed")
                    return

                alerts.success(f"Transposed {begin.value} to {end.value} in {num_changed} notes")
            case "transpose_key":
                key = self.questions.ask_existing_key(question="Which key to transpose?")
                if key is None:  # pragma: no cover
                    return

                num_changed = self.vault.transpose_metadata(
                    begin=begin,
                    end=end,
                    key=key,
                    location=self.vault.insert_location,
                )

                if num_changed == 0:
                    alerts.warning(f"No notes were changed")
                    return

                alerts.success(
                    f"Transposed key: `{key}` from {begin.value} to {end.value} in {num_changed} notes"
                )
            case "transpose_value":
                key = self.questions.ask_existing_key(question="Which key contains the value?")
                if key is None:  # pragma: no cover
                    return

                questions2 = Questions(vault=self.vault, key=key)
                value = questions2.ask_existing_value(question="Which value to transpose?")
                if value is None:  # pragma: no cover
                    return

                num_changed = self.vault.transpose_metadata(
                    begin=begin,
                    end=end,
                    key=key,
                    value=value,
                    location=self.vault.insert_location,
                )

                if num_changed == 0:
                    alerts.warning(f"No notes were changed")
                    return

                alerts.success(
                    f"Transposed key: `{key}:{value}` from {begin.value} to {end.value} in {num_changed} notes"
                )
            case _:
                return



"""Shared utilities."""

from obsidian_metadata._utils import alerts
from obsidian_metadata._utils.alerts import LoggerManager
from obsidian_metadata._utils.utilities import (
    clean_dictionary,
    clear_screen,
    dict_contains,
    dict_values_to_lists_strings,
    docstring_parameter,
    merge_dictionaries,
    remove_markdown_sections,
    version_callback,
)

__all__ = [
    "alerts",
    "clean_dictionary",
    "clear_screen",
    "dict_contains",
    "dict_values_to_lists_strings",
    "docstring_parameter",
    "LoggerManager",
    "merge_dictionaries",
    "remove_markdown_sections",
    "vault_validation",
    "version_callback",
]



"""Instantiate the configuration object."""


from pathlib import Path
from textwrap import dedent
from typing import Any

import questionary
import rich.repr
import tomlkit
import typer

from obsidian_metadata._utils import alerts
from obsidian_metadata._utils.alerts import logger as log


class ConfigQuestions:
    """Questions to ask the user when creating a configuration file."""

    @staticmethod
    def _validate_valid_dir(path: str) -> bool | str:
        """Validate a valid directory.

        Returns:
            bool | str: True if the path is valid, otherwise a string with the error message.
        """
        path_to_validate: Path = Path(path).expanduser().resolve()
        if not path_to_validate.exists():
            return f"Path does not exist: {path_to_validate}"
        if not path_to_validate.is_dir():
            return f"Path is not a directory: {path_to_validate}"

        return True

    @staticmethod
    def ask_for_vault_path() -> Path:  # pragma: no cover
        """Ask the user for the path to their vault.

        Returns:
            Path: The path to the vault.
        """
        vault_path = questionary.path(
            "Enter a path to Obsidian vault:",
            only_directories=True,
            validate=ConfigQuestions._validate_valid_dir,
        ).ask()
        if vault_path is None:
            raise typer.Exit(code=1)

        return Path(vault_path).expanduser().resolve()


@rich.repr.auto
class Config:
    """Representation of a configuration file."""

    def __init__(self, config_path: Path = None, vault_path: Path = None) -> None:
        if vault_path is None:
            self.config_path: Path = self._validate_config_path(Path(config_path))
            self.config: dict[str, Any] = self._load_config()

            if self.config == {}:
                log.error(f"Configuration file is empty: '{self.config_path}'")
                raise typer.Exit(code=1)
        else:
            self.config_path = None
            self.config = {
                "command_line_vault": {
                    "path": vault_path,
                    "exclude_paths": [".git", ".obsidian"],
                    "insert_location": "BOTTOM",
                }
            }

        try:
            self.vaults: list[VaultConfig] = [
                VaultConfig(vault_name=key, vault_config=self.config[key]) for key in self.config
            ]
        except TypeError as e:
            log.error(f"Configuration file is invalid: '{self.config_path}'")
            raise typer.Exit(code=1) from e

        log.debug(f"Loaded configuration from '{self.config_path}'")
        log.trace(self.config)

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Define rich representation of the Config object."""
        yield "config_path", self.config_path
        yield "vaults", self.vaults

    def _load_config(self) -> dict[str, Any]:
        """Load the configuration file."""
        try:
            with open(self.config_path, encoding="utf-8") as fp:
                return tomlkit.load(fp)
        except tomlkit.exceptions.TOMLKitError as e:
            alerts.error(f"Could not parse '{self.config_path}'")
            raise typer.Exit(code=1) from e

    def _validate_config_path(self, config_path: Path | None) -> Path:
        """Load the configuration path."""
        if config_path is None:
            config_path = Path(Path.home() / f".{__package__.split('.')[0]}.toml")

        if not config_path.exists():
            self._write_default_config(config_path)
            alerts.info(f"Created default configuration file at '{config_path}'")

        return config_path.expanduser().resolve()

    def _write_default_config(self, path_to_config: Path) -> None:
        """Write the default configuration file when no config file is found."""
        vault_path = ConfigQuestions.ask_for_vault_path()

        config_text = f"""\
        # Add another vault by replicating this section and changing the name
        ["Vault 1"] # Name of the vault.

            # Path to your obsidian vault
            path = "{vault_path}"

            # Folders within the vault to ignore when indexing metadata
            exclude_paths = [".git", ".obsidian"]

            # Location to add metadata. One of:
            #    TOP:            Directly after frontmatter.
            #    AFTER_TITLE:    After a header following frontmatter.
            #    BOTTOM:         The bottom of the note
            insert_location = "BOTTOM"
        """

        path_to_config.write_text(dedent(config_text))


@rich.repr.auto
class VaultConfig:
    """Representation of a vault configuration."""

    def __init__(self, vault_name: str, vault_config: dict) -> None:
        """Initialize the vault configuration."""
        self.name: str = vault_name
        self.config: dict = vault_config

        try:
            self.path = self._validate_vault_path(self.config["path"])

            Path(self.config["path"]).expanduser().resolve()
        except KeyError:
            self.path = None

        try:
            self.exclude_paths = self.config["exclude_paths"]
        except KeyError:
            self.exclude_paths = [".git", ".obsidian"]

        try:
            self.insert_location = self.config["insert_location"]
        except KeyError:
            self.insert_location = "BOTTOM"

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Define rich representation of a vault config."""
        yield "name", self.name
        yield "config", self.config
        yield "path", self.path
        yield "exclude_paths", self.exclude_paths

    def _validate_vault_path(self, vault_path: Path | None) -> Path:
        """Validate the vault path."""
        vault_path = Path(vault_path).expanduser().resolve()

        if not vault_path.exists():
            alerts.error(f"Vault path not found: '{vault_path}'")
            raise typer.Exit(code=1)

        if not vault_path.is_dir():
            alerts.error(f"Vault path is not a directory: '{vault_path}'")
            raise typer.Exit(code=1)

        return vault_path



"""Regexes for parsing frontmatter and note content."""

from dataclasses import dataclass

import regex as re
from regex import Pattern


@dataclass
class Patterns:
    """Regex patterns for parsing frontmatter and note content."""

    find_inline_tags: Pattern[str] = re.compile(
        r"""
        (?:^|[ \|_,;:\*\)\[\]\\\.]|(?<!\])\()   # Before tag is start of line or separator
        (?<!\/\/[\w\d_\.\(\)\/&_-]+)            # Before tag is not a link
        \#([^ \|,;:\*\(\)\[\]\\\.\n#&]+)        # Match tag until separator or end of line
        """,
        re.MULTILINE | re.X,
    )

    find_inline_metadata: Pattern[str] = re.compile(
        r"""                                    # First look for in-text key values
        (?:^\[| \[)                             # Find key with starting bracket
        ([-_\w\d\/\*\u263a-\U0001f999]+?)::[ ]? # Find key
        (.*?)\]                                 # Find value until closing bracket
        |                                       # Else look for key values at start of line
        (?:^|[^ \w\d]+| \[)                     # Any non-word or non-digit character
        ([-_\w\d\/\*\u263a-\U0001f9995]+?)::(?!\n)(?:[ ](?!\n))?  # Capture the key if not a new line
        (.*?)$                                  # Capture the value
        """,
        re.X | re.MULTILINE,
    )

    frontmatter_block: Pattern[str] = re.compile(r"^\s*(?P<frontmatter>---.*?---)", flags=re.DOTALL)
    frontmatt_block_strip_separators: Pattern[str] = re.compile(
        r"^\s*---(?P<frontmatter>.*?)---", flags=re.DOTALL
    )
    # This pattern will return a tuple of 4 values, two will be empty and will need to be stripped before processing further

    top_with_header: Pattern[str] = re.compile(
        r"""^\s*                                        # Start of note
        (?P<top>                                        # Capture the top of the note
            (---.*?---)?                                # Frontmatter, if it exists
            \s*                                         # Any whitespace
            (                                           # Full header, if it exists
                \#+[ ]                                  # Match start of any header level
                (                                       # Text of header
                    [\w\d]+                             # Word or digit
                    |                                   # Or
                    [\[\]\(\)\+\{\}\"'\-\.\/\*\$\| ]+   # Special characters
                    |                                   # Or
                    [\u263a-\U0001f999]+                # Emoji
                )+                                      # End of header text
            )?                                          # End of full header
        )                                               # End capture group
        """,
        flags=re.DOTALL | re.X,
    )

    validate_key_text: Pattern[str] = re.compile(r"[^-_\w\d\/\*\u263a-\U0001f999]")
    validate_tag_text: Pattern[str] = re.compile(r"[ \|,;:\*\(\)\[\]\\\.\n#&]")



"""Functions for asking questions to the user and validating responses.

This module contains wrappers around questionary to ask questions to the user and validate responses. Mocking questionary has proven very difficult. This functionality is separated from the main application logic to make it easier to test.

Progress towards testing questionary can be found on this issue:
https://github.com/tmbo/questionary/issues/35
"""
import re
from pathlib import Path
from typing import Any

import questionary
import typer

from obsidian_metadata.models.enums import InsertLocation, MetadataType
from obsidian_metadata.models.patterns import Patterns
from obsidian_metadata.models.vault import Vault

PATTERNS = Patterns()

# Reset the default style of the questionary prompts qmark
questionary.prompts.checkbox.DEFAULT_STYLE = questionary.Style([("qmark", "")])
questionary.prompts.common.DEFAULT_STYLE = questionary.Style([("qmark", "")])
questionary.prompts.confirm.DEFAULT_STYLE = questionary.Style([("qmark", "")])
questionary.prompts.confirm.DEFAULT_STYLE = questionary.Style([("qmark", "")])
questionary.prompts.path.DEFAULT_STYLE = questionary.Style([("qmark", "")])
questionary.prompts.select.DEFAULT_STYLE = questionary.Style([("qmark", "")])
questionary.prompts.text.DEFAULT_STYLE = questionary.Style([("qmark", "")])


class Questions:
    """Class for asking questions to the user and validating responses with questionary."""

    @staticmethod
    def ask_for_vault_path() -> Path:  # pragma: no cover
        """Ask the user for the path to their vault.

        Returns:
            Path: The path to the vault.
        """
        vault_path = questionary.path(
            "Enter a path to Obsidian vault:",
            only_directories=True,
            validate=Questions._validate_valid_dir,
        ).ask()
        if vault_path is None:
            raise typer.Exit(code=1)

        return Path(vault_path).expanduser().resolve()

    @staticmethod
    def _validate_valid_dir(path: str) -> bool | str:
        """Validate a valid directory.

        Returns:
            bool | str: True if the path is valid, otherwise a string with the error message.
        """
        path_to_validate: Path = Path(path).expanduser().resolve()
        if not path_to_validate.exists():
            return f"Path does not exist: {path_to_validate}"
        if not path_to_validate.is_dir():
            return f"Path is not a directory: {path_to_validate}"

        return True

    def __init__(self, vault: Vault = None, key: str = None) -> None:
        """Initialize the class.

        Args:
            vault_path (Path, optional): The path to the vault. Defaults to None.
            vault (Vault, optional): The vault object. Defaults to None.
            key (str, optional): The key to use when validating a key, value pair. Defaults to None.
        """
        self.style = questionary.Style(
            [
                ("qmark", "bold"),
                ("question", "bold"),
                ("separator", "fg:#808080"),
                ("answer", "fg:#FF9D00 bold"),
                ("instruction", "fg:#808080"),
                ("highlighted", "bold underline"),
                ("text", ""),
                ("pointer", "bold"),
            ]
        )
        self.vault = vault
        self.key = key

    def _validate_existing_inline_tag(self, text: str) -> bool | str:
        """Validate an existing inline tag.

        Returns:
            bool | str: True if the tag is valid, otherwise a string with the error message.
        """
        if len(text) < 1:
            return "Tag cannot be empty"

        if not self.vault.metadata.contains(area=MetadataType.TAGS, value=text):
            return f"'{text}' does not exist as a tag in the vault"

        return True

    def _validate_key_exists(self, text: str) -> bool | str:
        """Validate a valid key.

        Returns:
            bool | str: True if the key is valid, otherwise a string with the error message.
        """
        if len(text) < 1:
            return "Key cannot be empty"

        if not self.vault.metadata.contains(area=MetadataType.KEYS, key=text):
            return f"'{text}' does not exist as a key in the vault"

        return True

    def _validate_key_exists_regex(self, text: str) -> bool | str:
        """Validate a valid key.

        Returns:
            bool | str: True if the key is valid, otherwise a string with the error message.
        """
        if len(text) < 1:
            return "Key cannot be empty"

        try:
            re.compile(text)
        except re.error as error:
            return f"Invalid regex: {error}"

        if not self.vault.metadata.contains(area=MetadataType.KEYS, key=text, is_regex=True):
            return f"'{text}' does not exist as a key in the vault"

        return True

    def _validate_new_key(self, text: str) -> bool | str:
        """Validate the tag name.

        Args:
            text (str): The key name to validate.

        Returns:
            bool | str: True if the key is valid, otherwise a string with the error message.
        """
        if PATTERNS.validate_key_text.search(text) is not None:
            return "Key cannot contain spaces or special characters"

        if len(text) == 0:
            return "New key cannot be empty"

        return True

    def _validate_new_tag(self, text: str) -> bool | str:
        """Validate the tag name.

        Args:
            text (str): The tag name to validate.

        Returns:
            bool | str: True if the tag is valid, otherwise a string with the error message.
        """
        if PATTERNS.validate_tag_text.search(text) is not None:
            return "Tag cannot contain spaces or special characters"

        if len(text) == 0:
            return "New tag cannot be empty"

        return True

    def _validate_new_value(self, text: str) -> bool | str:
        """Validate a new value.

        Args:
            text (str): The value to validate.

        Returns:
            bool | str: True if the value is valid, otherwise a string with the error message.
        """
        if len(text) < 1:
            return "Value cannot be empty"

        if self.key is not None and self.vault.metadata.contains(
            area=MetadataType.ALL, key=self.key, value=text
        ):
            return f"{self.key}:{text} already exists"

        return True

    def _validate_number(self, text: str) -> bool | str:
        """Validate a number.

        Args:
            text (str): The number to validate.

        Returns:
            bool | str: True if the number is valid, otherwise a string with the error message.
        """
        if not text.isdigit():
            return "Must be an integer"

        return True

    def _validate_valid_vault_regex(self, text: str) -> bool | str:
        """Validate a valid regex.

        Returns:
            bool | str: True if the regex is valid, otherwise a string with the error message.
        """
        try:
            re.compile(text)
        except re.error as error:
            return f"Invalid regex: {error}"

        if self.vault is not None:
            for subdir in list(self.vault.vault_path.glob("**/*")):
                if re.search(text, str(subdir)):
                    return True
            return "Regex does not match paths in the vault"

        return True

    def _validate_value(self, text: str) -> bool | str:
        """Validate the value.

        Args:
            text (str): The value to validate.

        Returns:
            bool | str: True if the value is valid, otherwise a string with the error message.
        """
        if len(text) == 0:
            return True

        if self.key is not None and not self.vault.metadata.contains(
            area=MetadataType.ALL, key=self.key, value=text
        ):
            return f"{self.key}:{text} does not exist"

        return True

    def _validate_value_exists_regex(self, text: str) -> bool | str:
        """Validate the value.

        Args:
            text (str): The value to validate.

        Returns:
            bool | str: True if the value is valid, otherwise a string with the error message.
        """
        if len(text) < 1:
            return "Regex cannot be empty"

        try:
            re.compile(text)
        except re.error as error:
            return f"Invalid regex: {error}"

        if self.key is not None and not self.vault.metadata.contains(
            area=MetadataType.ALL, key=self.key, value=text, is_regex=True
        ):
            return f"No values in {self.key} match regex: {text}"

        return True

    def ask_application_main(self) -> str:  # pragma: no cover
        """List for the main application interface.

        Args:
            style (questionary.Style): The style to use for the question.

        Returns:
            str: The selected application.
        """
        return questionary.select(
            "What do you want to do?",
            choices=[
                {"name": "Vault Actions", "value": "vault_actions"},
                {"name": "Inspect Metadata", "value": "inspect_metadata"},
                {"name": "Filter Notes in Scope", "value": "filter_notes"},
                {"name": "Add Metadata", "value": "add_metadata"},
                {"name": "Rename Metadata", "value": "rename_metadata"},
                {"name": "Delete Metadata", "value": "delete_metadata"},
                {"name": "Transpose Metadata", "value": "transpose_metadata"},
                questionary.Separator("-------------------------------"),
                {"name": "Review Changes", "value": "review_changes"},
                {"name": "Commit Changes", "value": "commit_changes"},
                questionary.Separator("-------------------------------"),
                {"name": "Quit", "value": "abort"},
            ],
            use_shortcuts=False,
            style=self.style,
            qmark="INPUT    |",
        ).ask()

    def ask_area(self) -> MetadataType | str:  # pragma: no cover
        """Ask the user for the metadata area to work on.

        Returns:
            MetadataType: The metadata area to work on.
        """
        choices = []
        for metadata_type in MetadataType:
            choices.append({"name": metadata_type.value, "value": metadata_type})

        choices.append(questionary.Separator())  # type: ignore [arg-type]
        choices.append({"name": "Cancel", "value": "cancel"})
        return self.ask_selection(
            choices=choices,
            question="Select the type of metadata",
        )

    def ask_confirm(self, question: str, default: bool = True) -> bool:  # pragma: no cover
        """Ask the user to confirm an action.

        Args:
            question (str): The question to ask.
            default (bool, optional): The default value. Defaults to True.

        Returns:
            bool: True if the user confirms, otherwise False.
        """
        return questionary.confirm(
            question, default=default, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_existing_inline_tag(self, question: str = "Enter a tag") -> str:  # pragma: no cover
        """Ask the user for an existing inline tag."""
        return questionary.text(
            question,
            validate=self._validate_existing_inline_tag,
            style=self.style,
            qmark="INPUT    |",
        ).ask()

    def ask_existing_key(self, question: str = "Enter a key") -> str:  # pragma: no cover
        """Ask the user for a metadata key.

        Args:
            question (str, optional): The question to ask. Defaults to "Enter a key".

        Returns:
            str: A metadata key that exists in the vault.
        """
        return questionary.text(
            question, validate=self._validate_key_exists, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_existing_keys_regex(self, question: str = "Regex for keys") -> str:  # pragma: no cover
        """Ask the user for a regex for metadata keys.

        Args:
            question (str, optional): The question to ask. Defaults to "Regex for keys".

        Returns:
            str: A regex for metadata keys that exist in the vault.
        """
        return questionary.text(
            question, validate=self._validate_key_exists_regex, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_existing_value(self, question: str = "Enter a value") -> str:  # pragma: no cover
        """Ask the user for a metadata value.

        Args:
            question (str, optional): The question to ask. Defaults to "Enter a value".

        Returns:
            str: A metadata value.
        """
        return questionary.text(
            question, validate=self._validate_value, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_filter_path(self) -> str:  # pragma: no cover
        """Ask the user for the path to the filter file.

        Returns:
            str: The regex to use for filtering.
        """
        filter_path_regex = questionary.path(
            "Regex to filter the notes being processed by their path:",
            only_directories=False,
            validate=self._validate_valid_vault_regex,
            style=self.style,
            qmark="INPUT    |",
        ).ask()
        if filter_path_regex is None:
            raise typer.Exit(code=1)

        return filter_path_regex

    def ask_existing_value_regex(
        self, question: str = "Regex for values"
    ) -> str:  # pragma: no cover
        """Ask the user for a regex for metadata values.

        Args:
            question (str, optional): The question to ask. Defaults to "Regex for values".

        Returns:
            str: A regex for metadata values that exist in the vault.
        """
        return questionary.text(
            question,
            validate=self._validate_value_exists_regex,
            style=self.style,
            qmark="INPUT    |",
        ).ask()

    def ask_metadata_location(
        self, question: str = "Where in a note should we add metadata"
    ) -> InsertLocation:  # pragma: no cover
        """Ask the user for the location within a note to place new metadata.

        Returns:
            InsertLocation: The location within a note to place new metadata.
        """
        choices = []
        for metadata_location in InsertLocation:
            choices.append({"name": metadata_location.value, "value": metadata_location})

        return self.ask_selection(
            choices=choices,
            question="Select the location for the metadata",
        )

    def ask_new_key(self, question: str = "New key name") -> str:  # pragma: no cover
        """Ask the user for a new metadata key.

        Args:
            question (str, optional): The question to ask. Defaults to "New key name".

        Returns:
            str: A new metadata key.
        """
        return questionary.text(
            question, validate=self._validate_new_key, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_new_tag(self, question: str = "Enter a new tag") -> str:  # pragma: no cover
        """Ask the user for a new tag.

        Args:
            question (str, optional): The question to ask. Defaults to "Enter a new tag".
        """
        return questionary.text(
            question, validate=self._validate_new_tag, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_new_value(self, question: str = "New value") -> str:  # pragma: no cover
        """Ask the user for a new metadata value.

        Args:
            question (str, optional): The question to ask. Defaults to "New value".

        Returns:
            str: A new metadata value.
        """
        return questionary.text(
            question, validate=self._validate_new_value, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_number(self, question: str = "Enter a number") -> int:  # pragma: no cover
        """Ask the user for a number.

        Args:
            question (str, optional): The question to ask. Defaults to "Enter a number".

        Returns:
            int: A number.
        """
        return questionary.text(
            question, validate=self._validate_number, style=self.style, qmark="INPUT    |"
        ).ask()

    def ask_path(self, question: str = "Enter a path") -> str:  # pragma: no cover
        """Ask the user for a path.

        Args:
            question (str, optional): The question to ask. Defaults to "Enter a path".

        Returns:
            str: A path.
        """
        return questionary.path(question, style=self.style, qmark="INPUT    |").ask()

    def ask_selection(
        self, choices: list[Any], question: str = "Select an option"
    ) -> Any:  # pragma: no cover
        """Ask the user to select an item from a list.

        Args:
            question (str, optional): The question to ask. Defaults to "Select an option".
            choices (list[Any]): The list of choices.

        Returns:
            any: The selected item value.
        """
        choices.insert(0, questionary.Separator())
        return questionary.select(
            question,
            choices=choices,
            use_shortcuts=False,
            style=self.style,
            qmark="INPUT    |",
        ).ask()



"""Enum classes for the obsidian_metadata package."""

from enum import Enum


class MetadataType(Enum):
    """Enum class for the type of metadata."""

    FRONTMATTER = "Frontmatter"
    INLINE = "Inline Metadata"
    TAGS = "Inline Tags"
    KEYS = "Metadata Keys Only"
    ALL = "All Metadata"


class InsertLocation(Enum):
    """Location to add metadata to notes.

    TOP:            Directly after frontmatter.
    AFTER_TITLE:    After a header following frontmatter.
    BOTTOM:         The bottom of the note

    """

    TOP = "Top"
    AFTER_TITLE = "Header"
    BOTTOM = "Bottom"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['obsidian_metadata',
 'obsidian_metadata._config',
 'obsidian_metadata._utils',
 'obsidian_metadata.models']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.6.0,<0.7.0',
 'questionary>=1.10.0,<2.0.0',
 'regex>=2022.10.31,<2023.0.0',
 'rich>=13.2.0,<14.0.0',
 'ruamel-yaml>=0.17.21,<0.18.0',
 'shellingham>=1.4.0,<2.0.0',
 'tomlkit>=0.11.6,<0.12.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['obsidian-metadata = obsidian_metadata.cli:app']}

setup_kwargs = {
    'name': 'obsidian-metadata',
    'version': '0.6.0',
    'description': 'Make batch updates to Obsidian metadata',
    'long_description': '[![PyPI version](https://badge.fury.io/py/obsidian-metadata.svg)](https://badge.fury.io/py/obsidian-metadata) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/obsidian-metadata) [![Python Code Checker](https://github.com/natelandau/obsidian-metadata/actions/workflows/automated-tests.yml/badge.svg)](https://github.com/natelandau/obsidian-metadata/actions/workflows/automated-tests.yml) [![codecov](https://codecov.io/gh/natelandau/obsidian-metadata/branch/main/graph/badge.svg?token=3F2R43SSX4)](https://codecov.io/gh/natelandau/obsidian-metadata)\n\n# obsidian-metadata\n\nA script to make batch updates to metadata in an Obsidian vault. No changes are\nmade to the Vault until they are explicitly committed.\n\n[![asciicast](https://asciinema.org/a/555789.svg)](https://asciinema.org/a/555789)\n\n## Important Disclaimer\n\n**It is strongly recommended that you back up your vault prior to committing changes.** This script makes changes directly to the markdown files in your vault. Once the changes are committed, there is no ability to recreate the original information unless you have a backup. Follow the instructions in the script to create a backup of your vault if needed. The author of this script is not responsible for any data loss that may occur. Use at your own risk.\n\n## Install\n\nRequires Python v3.10 or above.\n\n```bash\npip install obsidian-metadata\n```\n\n## Usage\n\n### CLI Commands\n\n-   `--config-file`: Specify a custom configuration file location\n-   `--dry-run`: Make no destructive changes\n-   `--export-csv`: Specify a path and create a CSV export of all metadata\n-   `--export-json`: Specify a path and create a JSON export of all metadata\n-   `--help`: Shows interactive help and exits\n-   `--log-file`: Specify a log file location\n-   `--log-to-file`: Will log to a file\n-   `--vault-path`: Specify a path to an Obsidian Vault\n-   `--verbose`: Set verbosity level (0=WARN, 1=INFO, 2=DEBUG, 3=TRACE)\n-   `--version`: Prints the version number and exits\n\n### Running the script\n\nOnce installed, run `obsidian-metadata` in your terminal to enter an interactive menu of sub-commands.\n\n**Vault Actions**\n\n-   Backup: Create a backup of the vault.\n-   Delete Backup: Delete a backup of the vault.\n\n**Inspect Metadata**\n\n-   **View all metadata in the vault**\n-   View all **frontmatter**\n-   View all **inline metadata**\n-   View all **inline tags**\n-   **Export all metadata to CSV or JSON file**\n\n**Filter Notes in Scope**: Limit the scope of notes to be processed with one or more filters.\n\n-   **Path filter (regex)**: Limit scope based on the path or filename\n-   **Metadata filter**: Limit scope based on a key or key/value pair\n-   **Tag filter**: Limit scope based on an in-text tag\n-   **List and clear filters**: List all current filters and clear one or all\n-   **List notes in scope**: List notes that will be processed.\n\n**Add Metadata**: Add new metadata to your vault.\n\n-   **Add new metadata to the frontmatter**\n-   **Add new inline metadata** - Set `insert_location` in the config to control where the new metadata is inserted. (Default: Bottom)\n-   **Add new inline tag** - Set `insert_location` in the config to control where the new tag is inserted. (Default: Bottom)\n\n**Rename Metadata**: Rename either a key and all associated values, a specific value within a key. or an in-text tag.\n\n-   **Rename a key**\n-   **Rename a value**\n-   **Rename an inline tag**\n\n**Delete Metadata**: Delete either a key and all associated values, or a specific value.\n\n-   **Delete a key and associated values**\n-   **Delete a value from a key**\n-   **Delete an inline tag**\n\n**Transpose Metadata**: Move metadata from inline to frontmatter or the reverse.\n\n-   **Transpose all metadata** - Moves all frontmatter to inline metadata, or the reverse\n-   **Transpose key** - Transposes a specific key and all it\'s values\n-   **Transpose value**- Transpose a specific key:value pair\n\n**Review Changes**: Prior to committing changes, review all changes that will be made.\n\n-   **View a diff of the changes** that will be made\n\n**Commit Changes**: Write the changes to disk. This step is not undoable.\n\n-   **Commit changes to the vault**\n\n### Configuration\n\n`obsidian-metadata` requires a configuration file at `~/.obsidian_metadata.toml`. On first run, this file will be created. You can specify a new location for the configuration file with the `--config-file` option.\n\nTo add additional vaults, copy the default section and add the appropriate information. The script will prompt you to select a vault if multiple exist in the configuration file\n\nBelow is an example with two vaults.\n\n```toml\n["Vault One"] # Name of the vault.\n\n    # Path to your obsidian vault\n    path = "/path/to/vault"\n\n    # Folders within the vault to ignore when indexing metadata\n    exclude_paths = [".git", ".obsidian"]\n\n    # Location to add metadata. One of:\n    #    TOP:            Directly after frontmatter.\n    #    AFTER_TITLE:    After a header following frontmatter.\n    #    BOTTOM:         The bottom of the note\n    insert_location = "BOTTOM"\n\n["Vault Two"]\n    path = "/path/to/second_vault"\n    exclude_paths = [".git", ".obsidian", "daily_notes"]\n    insert_location = "AFTER_TITLE"\n```\n\nTo bypass the configuration file and specify a vault to use at runtime use the `--vault-path` option.\n\n# Contributing\n\n## Setup: Once per project\n\nThere are two ways to contribute to this project.\n\n### 1. Containerized development\n\n1. Clone this repository. `git clone https://github.com/natelandau/obsidian-metadata`\n2. Open the repository in Visual Studio Code\n3. Start the [Dev Container](https://code.visualstudio.com/docs/remote/containers). Run <kbd>Ctrl/⌘</kbd> + <kbd>⇧</kbd> + <kbd>P</kbd> → _Remote-Containers: Reopen in Container_.\n4. Run `poetry env info -p` to find the PATH to the Python interpreter if needed by VSCode.\n\n### 2. Local development\n\n1. Install Python 3.10 and [Poetry](https://python-poetry.org)\n2. Clone this repository. `git clone https://github.com/natelandau/obsidian-metadata`\n3. Install the Poetry environment with `poetry install`.\n4. Activate your Poetry environment with `poetry shell`.\n5. Install the pre-commit hooks with `pre-commit install --install-hooks`.\n\n## Developing\n\n-   This project follows the [Conventional Commits](https://www.conventionalcommits.org/) standard to automate [Semantic Versioning](https://semver.org/) and [Keep A Changelog](https://keepachangelog.com/) with [Commitizen](https://github.com/commitizen-tools/commitizen).\n    -   When you\'re ready to commit changes run `cz c`\n-   Run `poe` from within the development environment to print a list of [Poe the Poet](https://github.com/nat-n/poethepoet) tasks available to run on this project. Common commands:\n    -   `poe lint` runs all linters\n    -   `poe test` runs all tests with Pytest\n-   Run `poetry add {package}` from within the development environment to install a run time dependency and add it to `pyproject.toml` and `poetry.lock`.\n-   Run `poetry remove {package}` from within the development environment to uninstall a run time dependency and remove it from `pyproject.toml` and `poetry.lock`.\n-   Run `poetry update` from within the development environment to upgrade all dependencies to the latest versions allowed by `pyproject.toml`.\n',
    'author': 'Nate Landau',
    'author_email': 'github@natenate.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/natelandau/obsidian-metadata',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
